"""
Template Version Manager API
"""
from core_main_app.components.template import api as template_api
from core_main_app.components.template_version_manager.models import TemplateVersionManager
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.components.version_manager.utils import get_latest_version_name, get_version_name


def insert(template_version_manager, template):
    """Add a version to a template version manager.

    Args:
        template_version_manager:
        template:

    Returns:

    """
    # save the template in database
    template_api.upsert(template)
    try:
        # insert the initial template in the version manager
        version_manager_api.insert_version(template_version_manager, template)
        # insert the version manager in database
        version_manager_api.upsert(template_version_manager)
        # get template display name
        display_name = get_latest_version_name(template_version_manager)
        # update saved template
        template_api.set_display_name(template, display_name)
        # return version manager
        return template_version_manager
    except Exception as e:
        template_api.delete(template)
        raise e


def edit_title(template_version_manager, title):
    """Edit template version manager title.

    Args:
        template_version_manager:
        title:

    Returns:

    """
    # set version template manager title
    template_version_manager.title = title
    # save template version manager
    template_version_manager.save_template_version_manager()
    # update templates display names
    update_templates_display_name(template_version_manager)


def update_templates_display_name(template_version_manager):
    """Update templates display name.

    Args:
        template_version_manager:

    Returns:

    """
    # Iterate versions
    for i in range(0, len(template_version_manager.versions)):
        # get template id from list of versions
        template_id = template_version_manager.versions[i]
        # get template from template id
        template = template_api.get(template_id)
        # get display name for the template
        display_name = get_version_name(template_version_manager.title, i+1)
        # update template's display name
        template_api.set_display_name(template, display_name)


def get_by_id(template_version_manager_id):
    """Get a template version manager by its id.

    Args:
        template_version_manager_id: Id.

    Returns:

    """
    return TemplateVersionManager.get_by_id(template_version_manager_id)


def get_global_version_managers(_cls=True):
    """Get all global version managers of a template.

    Returns:
        _cls:
    """
    return TemplateVersionManager.get_global_version_managers(_cls)


def get_active_global_version_manager(_cls=True):
    """ Return all active Version Managers with user set to None.

    Returns:

    """
    return TemplateVersionManager.get_active_global_version_manager(_cls)


def get_disable_global_version_manager(_cls=True):
    """ Return all disabled Version Managers with user set to None.

    Returns:

    """
    return TemplateVersionManager.get_disable_global_version_manager(_cls)


def get_active_version_manager_by_user_id(user_id, _cls=True):
    """ Return all active Version Managers with given user id.

    Returns:

    """
    return TemplateVersionManager.get_active_version_manager_by_user_id(user_id, _cls)


def get_disable_version_manager_by_user_id(user_id, _cls=True):
    """ Return all disabled Version Managers with given user id.

    Returns:

    """
    return TemplateVersionManager.get_disable_version_manager_by_user_id(user_id, _cls)


def get_version_number(template_version_manager, template):
    """ Return version number from version id.

    Args:
        template_version_manager:
        template:

    Returns:
        Version number

    """
    return template_version_manager.get_version_number(template)


def get_by_version_id(version_id):
    """Get the template version manager containing the given version id.

    Args:
        version_id: version id.

    Returns:
        template version manager.

    """
    return TemplateVersionManager.get_by_version_id(version_id)


def get_all_by_version_ids(version_ids):
    """Get all template version managers by a list of version ids.

    Args:
        version_ids: list of version ids.

    Returns:
        List of template version managers.

    """
    return TemplateVersionManager.get_all_by_version_ids(version_ids)


def get_all_version_manager_except_user_id(user_id, _cls=True):
    """ Return all  Version Managers of all users except user with given user id.

    Returns:

    """
    return TemplateVersionManager.get_all_version_manager_except_user_id(user_id, _cls)


def get_all_by_user_id(user_id, _cls=True):
    """ Return all Template Version Managers with given user id.

    Returns:

    """
    return TemplateVersionManager.get_all_version_manager_by_user_id(user_id, _cls)


def get_all(_cls=True):
    """ Return all Template Version Managers.

    Returns:

    """
    return TemplateVersionManager.get_all_version_manager(_cls)



""" Web page model
"""
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors

from core_main_app.commons import exceptions
from core_main_app.commons.enums import WEB_PAGE_TYPES


class WebPage(Document):
    """ Represents a WebPage
    """
    type = fields.IntField()
    content = fields.StringField()

    @staticmethod
    def get_by_type(page_type):
        """ Get a WebPage given its type

            Parameters:
                page_type (str): page type

            Returns:
                Web Page corresponding to the given type
        """
        try:
            return WebPage.objects.get(type=WEB_PAGE_TYPES[page_type])
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    @staticmethod
    def delete_by_type(page_type_key):
        """ Delete all WebPage with the given type key

        Args:
            page_type_key (int): page type key

        Returns:

        """
        return WebPage.objects(type=page_type_key).delete()



"""Units tests for User rest api
"""
from django.test import SimpleTestCase
from mock.mock import patch

from core_main_app.rest.user import views as user_rest_views
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock
from core_main_app.components.user import api as user_api


class TestGetAllUserList(SimpleTestCase):
    def setUp(self):
        super(TestGetAllUserList, self).setUp()
        self.data = None

    @patch.object(user_api, 'get_all_users')
    def test_get_all_returns_list(self, get_all_users):
        # Arrange
        user = create_mock_user('0', is_staff=True)

        # Act
        response = RequestMock.do_request_get(user_rest_views.UserList.as_view(), user, self.data)

        # Assert
        excepted_result = []
        self.assertEqual(response.data, excepted_result)



""" Data API
"""
import datetime

import pytz

import core_main_app.access_control.api
import core_main_app.components.workspace.access_control
from core_main_app.access_control import api as access_control_api
from core_main_app.access_control.decorators import access_control
from core_main_app.commons import exceptions as exceptions
from core_main_app.components.data import access_control as data_api_access_control
from core_main_app.components.data.models import Data
from core_main_app.components.workspace import api as workspace_api
from core_main_app.utils.xml import validate_xml_data
from core_main_app.settings import DATA_SORTING_FIELDS
from xml_utils.xsd_tree.xsd_tree import XSDTree


@access_control(access_control_api.can_read_or_write_in_workspace)
def get_all_by_workspace(workspace, user, order_by_field=DATA_SORTING_FIELDS):
    """ Get all data that belong to the workspace.

    Args:
        workspace:
        order_by_field:

    Returns:

    """
    return Data.get_all_by_workspace(workspace, order_by_field)


@access_control(data_api_access_control.can_read_list_data_id)
def get_by_id_list(list_data_id, user, order_by_field=DATA_SORTING_FIELDS):
    """ Return a list of data object with the given list id.

        Parameters:
            list_data_id:
            user:
            order_by_field:

        Returns: data object
    """
    return Data.get_all_by_id_list(list_data_id, order_by_field)


@access_control(core_main_app.access_control.api.can_read_id)
def get_by_id(data_id, user):
    """ Return data object with the given id.

        Parameters:
            data_id:
            user:

        Returns: data object
    """
    return Data.get_by_id(data_id)


@access_control(access_control_api.has_perm_administration)
def get_all(user, order_by_field=DATA_SORTING_FIELDS):
    """ Get all the data if superuser. Raise exception otherwise.

    Parameters:
            user:
            order_by_field: Order by field.

    Returns: data collection
    """
    return Data.get_all(order_by_field)


def get_all_accessible_by_user(user, order_by_field=DATA_SORTING_FIELDS):
    """ Return all data accessible by a user.

        Parameters:
            user:
            order_by_field:

        Returns: data collection
    """

    read_workspaces = workspace_api.get_all_workspaces_with_read_access_by_user(user)
    write_workspaces = workspace_api.get_all_workspaces_with_write_access_by_user(user)
    user_accessible_workspaces = list(set().union(read_workspaces, write_workspaces))

    return Data.get_all_by_user_and_workspace(user.id, user_accessible_workspaces, order_by_field)


def get_all_by_user(user, order_by_field=DATA_SORTING_FIELDS):
    """ Return all data owned by a user.

        Parameters:
            user:
            order_by_field: Order by field.

        Returns: data collection
    """
    return Data.get_all_by_user_id(str(user.id), order_by_field)


@access_control(core_main_app.access_control.api.can_read)
def get_all_except_user(user, order_by_field=DATA_SORTING_FIELDS):
    """ Return all data which are not created by the user.

        Parameters:
             user:
             order_by_field:

        Returns: data collection
    """
    return Data.get_all_except_user_id(str(user.id), order_by_field)


@access_control(core_main_app.access_control.api.can_write)
def upsert(data, user):
    """ Save or update the data.

    Args:
        data:
        user:

    Returns:

    """
    if data.xml_content is None:
        raise exceptions.ApiError("Unable to save data: xml_content field is not set.")

    data.last_modification_date = datetime.datetime.now(pytz.utc)
    check_xml_file_is_valid(data)
    return data.convert_and_save()


def check_xml_file_is_valid(data):
    """ Check if xml data is valid against a given schema.

    Args:
        data:

    Returns:

    """
    template = data.template

    try:
        xml_tree = XSDTree.build_tree(data.xml_content)
    except Exception as e:
        raise exceptions.XMLError(str(e))

    try:
        xsd_tree = XSDTree.build_tree(template.content)
    except Exception as e:
        raise exceptions.XSDError(str(e))

    error = validate_xml_data(xsd_tree, xml_tree)
    if error is not None:
        raise exceptions.XMLError(error)
    else:
        return True


@access_control(data_api_access_control.can_read_data_query)
def execute_query(query, user, order_by_field=DATA_SORTING_FIELDS):
    """Execute a query on the Data collection.

    Args:
        query:
        user:
        order_by_field:

    Returns:

    """
    return Data.execute_query(query, order_by_field)


@access_control(core_main_app.access_control.api.can_write)
def delete(data, user):
    """ Delete a data.

    Args:
        data:
        user:

    Returns:

    """
    data.delete()


@access_control(access_control_api.can_change_owner)
def change_owner(data, new_user, user):
    """ Change data's owner.

    Args:
        data:
        user:
        new_user:

    Returns:
    """
    # FIXME: user can transfer data to anybody, too permissive
    data.user_id = str(new_user.id)
    data.save()


def is_data_public(data):
    """ Is data public.

    Args:
        data:

    Returns:
    """
    return workspace_api.is_workspace_public(data.workspace) if data.workspace is not None else False


@access_control(data_api_access_control.can_read_aggregate_query)
def aggregate(pipeline, user):
    """Execute an aggregate on the Data collection.

    Args:
        pipeline:
        user:

    Returns:

    """
    return Data.aggregate(pipeline)


@access_control(data_api_access_control.can_write_data_workspace)
def assign(data, workspace, user):
    """ Assign blob to a workspace.

    Args:
        data:
        workspace:
        user:

    Returns:

    """
    data.workspace = workspace
    return data.save()



""" REST Views for XSLT Transformation
"""
from django.http import Http404
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated, IsAdminUser
from rest_framework.response import Response
from rest_framework.views import APIView

import core_main_app.components.xsl_transformation.api as xsl_api
from core_main_app.commons import exceptions
from core_main_app.rest.xsl_transformation.serializers import XslTransformationSerializer, TransformSerializer


class XslTransformationList(APIView):
    """ List, create XSL document
    """

    permission_classes = (IsAdminUser,)

    def get(self, request):
        """ Get all XSL document

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of XSL document
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            xsl_object_list = xsl_api.get_all()
            # Serialize object
            return_value = XslTransformationSerializer(xsl_object_list, many=True)
            # Return response
            return Response(return_value.data)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def post(self, request):
        """ Save an XSL document

        Parameters:

            {
                "name": "instance_name",
                "filename": "url",
                "content": "<content />",
            }

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Created XSL document
            - code: 400
              content: Validation error
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializer
            xsl_serializer = XslTransformationSerializer(data=request.data)
            # Validate xsl
            xsl_serializer.is_valid(True)
            # save or update the object
            xsl_serializer.save()
            return Response(xsl_serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class XslTransformationDetail(APIView):
    """" Get, delete, patch an XSL document
    """

    permission_classes = (IsAdminUser,)

    def get_object(self, pk):
        """ Get XSL document from db

        Args:

            pk: ObjectId

        Returns:

            XSL document
        """
        try:
            return xsl_api.get_by_id(pk)
        except exceptions.DoesNotExist:
            raise Http404

    def get(self, request, pk):
        """ Retrieve XSLT

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: XSL document
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            xsl_object = self.get_object(pk)
            # Serialize object
            return_value = XslTransformationSerializer(xsl_object)
            # Return response
            return Response(return_value.data)
        except Http404:
            content = {'message': 'Xsl not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def delete(self, request, pk):
        """ Delete an xsl document

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 204
              content: Deletion succeed
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            xsl_object = self.get_object(pk)
            # Remove the instance
            xsl_api.delete(xsl_object)
            # Return response
            return Response(status=status.HTTP_204_NO_CONTENT)
        except Http404:
            content = {'message': 'Xsl not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def patch(self, request, pk):
        """ Update xsl

        Parameters:

            {
                "name": "instance_name",
                "filename": "url",
                "content": "<content />",
            }

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: Updated XSL document
            - code: 400
              content: Validation error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            xsl_object = self.get_object(pk)
            # Build serializer
            xsl_serializer = XslTransformationSerializer(instance=xsl_object,
                                                         data=request.data,
                                                         partial=True)
            # Validate xsl
            xsl_serializer.is_valid(True)
            # Save xsl
            xsl_serializer.save()
            return Response(xsl_serializer.data, status=status.HTTP_200_OK)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Http404:
            content = {'message': 'Xsl not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class XslTransformationTransform(APIView):
    """ Transform XML using a specific Xslt
    """

    permission_classes = (IsAuthenticated,)

    def post(self, request):
        """ Transform XML using a specific Xslt

        Parameters:

            {
                "xml_content": "<xml />",
                "xslt_name": "name"
            }

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: transformed xml content
            - code: 400
              content: Validation error
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializer
            serializer = TransformSerializer(data=request.data)
            # Validate data
            serializer.is_valid(True)
            # transform
            return_value = xsl_api.xsl_transform(**serializer.validated_data)
            return Response(return_value, status=status.HTTP_200_OK)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)



""" Core Exceptions
"""


class CoreError(Exception):
    """ Exception raised by the Core.
    """
    def __init__(self, message):
        self.message = message


class ApiError(Exception):
    """ Exception raised by the API.
    """
    def __init__(self, message):
        self.message = message


class ModelError(Exception):
    """ Generic exception for the model.
    """
    def __init__(self, message):
        self.message = message


class RestApiError(Exception):
    """ Exception raised by the the REST API.
    """
    def __init__(self, message):
        self.message = message


class DoesNotExist(Exception):
    """ Exception raised when an object does not exist.
    """
    def __init__(self, message):
        self.message = message


class NotUniqueError(Exception):
    """ Exception raised when an object is not unique.
    """
    def __init__(self, message):
        self.message = message


class XMLError(Exception):
    """ Exception raised by XML validation.
    """
    def __init__(self, message):
        self.message = message


class XSDError(Exception):
    """ Exception raised by XSD validation.
    """
    def __init__(self, message):
        self.message = message


class LockError(Exception):
    """ Exception raised when an object is locked.
    """

    def __init__(self, message):
        self.message = message


class PaginationError(Exception):
    """ Exception raised when an error occurs during pagination.
    """
    def __init__(self, message):
        self.message = message


class BlobDownloaderUrlParseError(Exception):
    """ Exception raised when an error occurs during url parse.
    """
    def __init__(self, message):
        self.message = message


class BlobDownloaderError(Exception):
    """ Exception raised when an error occurs during blob download.
    """
    def __init__(self, message):
        self.message = message


class SSLError(Exception):
    """ Exception raised when an error occurs during SSL configuration.
    """
    def __init__(self, message):
        self.message = message



""" Templatetags to get JSON date format and display according to parsing directives
"""
import logging

import dateutil.parser

from django import template
from django.template.defaultfilters import stringfilter
from django.utils import formats

logger = logging.getLogger(__name__)
register = template.Library()


@register.filter(name="json_date")
@stringfilter
def json_date(value, directive):
    """ get JSON date string input and return a formatted date string according to the parsing directive

    Args:
        value: input JSON date string
        directive: standard django date format directive

    See: https://docs.djangoproject.com/fr/3.0/ref/templates/builtins/#date
    Returns: formatted string
    """
    formatted_date = value
    try:
        # parse the JSON date
        input_datetime = dateutil.parser.parse(formatted_date)
        # format the date thanks to the directive for UI display
        formatted_date = formats.date_format(input_datetime, directive)
        return formatted_date
    except Exception as e:
        logger.error(str(e))
        return formatted_date



""" Common Validators
"""
import os
import re

from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ungettext
from django.contrib.auth.password_validation import validate_password as django_validate_password
from django.contrib.auth.password_validation import MinimumLengthValidator, \
    UserAttributeSimilarityValidator, CommonPasswordValidator

from core_main_app.settings import PASSWORD_MIN_LENGTH, PASSWORD_MIN_UPPERCASE_LETTERS, PASSWORD_MIN_LOWERCASE_LETTERS, \
    PASSWORD_MIN_NUMBERS, PASSWORD_MIN_SYMBOLS, PASSWORD_MIN_LETTERS, PASSWORD_MAX_OCCURRENCE


@deconstructible
class BlankSpacesValidator(object):
    def __call__(self, value):
        value = force_text(value)
        if len(value.strip()) == 0:
            raise ValidationError(
                _('This field should not be empty.'),
            )


@deconstructible
class ExtensionValidator(object):
    def __init__(self, valid_extensions=list()):
        self.valid_extensions = valid_extensions

    def __call__(self, value):
        ext = os.path.splitext(value.name)[1]
        if not ext.lower() in self.valid_extensions:
            raise ValidationError(
                _('Unsupported file extension.'),
            )


class UpperCaseLetterCountValidator(object):
    """ Counts the occurrences of uppercase letters and raises a :class:`~django.core.exceptions.ValidationError` if the count
    is less than :func:`~UpperCaseLetterCountValidator.get_min_count`.
    """

    def __init__(self, min_uppercase_letters=0):
        self.min_uppercase_letters = min_uppercase_letters

    def validate(self, password, user=None):
        if sum(1 for char in password if char.isupper()) < self.get_min_count():
            raise ValidationError(self.get_help_text())

    def get_help_text(self):
        """ Returns this validator's error message.
        """
        msg = ungettext("The new password must contain %d or more uppercase letter.",
                        "The new password must contain %d or more uppercase letters.",
                        self.get_min_count()) % self.get_min_count()
        return msg

    def get_min_count(self):
        """ Returns: Min uppercase letters
        """
        return self.min_uppercase_letters


class LowerCaseLetterCountValidator(object):
    """ Counts the occurrences of lowercase letters and raises a :class:`~django.core.exceptions.ValidationError` if the count
    is less than :func:`~LowerCaseLetterCountValidator.get_min_count`.
    """

    def __init__(self, min_lowercase_letters=0):
        self.min_lowercase_letters = min_lowercase_letters

    def validate(self, password, user=None):
        if sum(1 for char in password if char.islower()) < self.get_min_count():
            raise ValidationError(self.get_help_text())

    def get_help_text(self):
        """ Returns this validator's error message.
        """
        msg = ungettext("The new password must contain %d or more lowercase letter.",
                        "The new password must contain %d or more lowercase letters.",
                        self.get_min_count()) % self.get_min_count()
        return msg

    def get_min_count(self):
        """ Returns: Min lowercase letters
        """
        return self.min_lowercase_letters


class AlphabeticCharCountValidator(object):
    """ Counts the occurrences of letters and raises a :class:`~django.core.exceptions.ValidationError` if the count
    is less than :func:`~AlphabeticCharCountValidator.get_min_count`.
    """

    def __init__(self, min_alphabetic_letters=0):
        self.min_alphabetic_letters = min_alphabetic_letters

    def validate(self, password, user=None):
        if sum(1 for char in password if self.is_alpha(char)) < self.get_min_count():
            raise ValidationError(self.get_help_text())

    def get_help_text(self):
        """ Returns this validator's error message.
        """
        msg = ungettext("The new password must contain %d or more alphabetic letter.",
                        "The new password must contain %d or more alphabetic letters.",
                        self.get_min_count()) % self.get_min_count()
        return msg

    def is_alpha(self, char):
        regex = r"[0-9a-zA-Z]"
        res = re.findall(regex, char)
        return len(res) > 0

    def get_min_count(self):
        """ Returns: Min alphabetic letters
        """
        return self.min_alphabetic_letters


class MaxOccurrenceCountValidator(object):
    """ Counts the occurrences of same letters and raises a :class:`~django.core.exceptions.ValidationError` if the count
    is more than :func:`~MaxOccurrenceCountValidator.get_min_count`.
    """

    def __init__(self, max_occurrence=None):
        self.max_occurrence = max_occurrence

    def validate(self, password, user=None):
        if self.max_occurrence and self.max_occurrence > 0:
            for current_char in password:
                if sum(1 for char in password if char == current_char) >= self.get_max_count():
                    raise ValidationError(self.get_help_text())

    def get_help_text(self):
        """ Returns this validator's error message.
        """
        msg = ungettext("The new password must contain less than %d time the same letter.",
                        "The new password must contain less than %d times the same letter.",
                        self.get_max_count()) % self.get_max_count()
        return msg

    def get_max_count(self):
        """ Returns: Max occurrence
        """
        return self.max_occurrence


class NonAlphanumericCountValidator(object):
    """ Counts the occurrences of Non-Alphanumeric and raises a :class:`~django.core.exceptions.ValidationError` if the count
    is less than :func:`~NonAlphanumericCountValidator.get_min_count`.
    """
    regex = r"[^0-9a-zA-Z]"

    def __init__(self, min_nonalphanumeric_letters=0):
        self.min_nonalphanumeric_letters = min_nonalphanumeric_letters

    def validate(self, password, user=None):
        res = re.findall(self.regex, password)
        if len(res) < self.get_min_count():
            raise ValidationError(self.get_help_text())

    def get_help_text(self):
        """ Returns this validator's error message.
        """
        msg = ungettext("The new password must contain %d or more Non-Alphanumeric letter.",
                        "The new password must contain %d or more Non-Alphanumeric letters.",
                        self.get_min_count()) % self.get_min_count()
        return msg

    def get_min_count(self):
        """ Returns: Min Non-Alphanumeric letters
        """
        return self.min_nonalphanumeric_letters


class DigitsCountValidator(object):
    """ Counts the occurrences of digits and raises a :class:`~django.core.exceptions.ValidationError` if the count
    is less than :func:`~DigitsCountValidator.get_min_count`.
    """
    regex = r"[0-9]"

    def __init__(self, min_digits_letters=0):
        self.min_digits_letters = min_digits_letters

    def validate(self, password, user=None):
        res = re.findall(self.regex, password)
        if len(res) < self.get_min_count():
            raise ValidationError(self.get_help_text())

    def get_help_text(self):
        """ Returns this validator's error message.
        """
        msg = ungettext("The new password must contain %d or more digits.",
                        "The new password must contain %d or more digit.",
                        self.get_min_count()) % self.get_min_count()
        return msg

    def get_min_count(self):
        """ Returns: Min Non-Alphanumeric letters
        """
        return self.min_digits_letters


def validate_password(password):
    """ Validate the given password, throw a ValidationError if at least on validator fail
    :param password:
    :return:
    """
    django_validate_password(password, password_validators=
    [MinimumLengthValidator(min_length=PASSWORD_MIN_LENGTH),
     UserAttributeSimilarityValidator(),
     CommonPasswordValidator(),
     UpperCaseLetterCountValidator(min_uppercase_letters=PASSWORD_MIN_UPPERCASE_LETTERS),
     LowerCaseLetterCountValidator(min_lowercase_letters=PASSWORD_MIN_LOWERCASE_LETTERS),
     NonAlphanumericCountValidator(min_nonalphanumeric_letters=PASSWORD_MIN_SYMBOLS),
     AlphabeticCharCountValidator(min_alphabetic_letters=PASSWORD_MIN_LETTERS),
     DigitsCountValidator(min_digits_letters=PASSWORD_MIN_NUMBERS),
     MaxOccurrenceCountValidator(
         max_occurrence=PASSWORD_MAX_OCCURRENCE)])


""" Set of functions to define the rules for access control
"""
import logging

import core_main_app.permissions.rights as rights
from core_main_app.access_control.api import has_perm_publish, check_can_read_list, can_write_in_workspace
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.components.workspace import api as workspace_api
from core_main_app.settings import CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT, VERIFY_DATA_ACCESS
from core_main_app.utils.labels import get_data_label
from core_main_app.utils.raw_query.mongo_raw_query import add_access_criteria, add_aggregate_access_criteria
from core_main_app.settings import DATA_SORTING_FIELDS

logger = logging.getLogger(__name__)


def has_perm_publish_data(user):
    """ Does the user have the permission to publish a data.

    Args:
        user

    Returns
    """
    has_perm_publish(user, rights.publish_data)


def can_read_list_data_id(func, list_data_id, user):
    """ Can read list of data.

    Args:
        func:
        list_data_id:
        user:

    Returns:

    """
    if user.is_superuser:
        return func(list_data_id, user)

    list_data = func(list_data_id, user)
    check_can_read_list(list_data, user)

    return list_data


def can_read_data_query(func, query, user, order_by_field=DATA_SORTING_FIELDS):
    """ Can read a data, given a query.

    Args:
        func:
        query:
        user:
        order_by_field:

    Returns:

    """
    if user.is_superuser:
        return func(query, user, order_by_field)

    # update the query
    query = _update_can_read_query(query, user)
    # get list of data
    data_list = func(query, user, order_by_field)
    # TODO: check if necessary because it is time consuming (checking that user has access to list of returned data)
    # check that user can access the list of data
    if VERIFY_DATA_ACCESS:
        check_can_read_list(data_list, user)
    return data_list


def can_read_aggregate_query(func, query, user):
    """ Can read a data, given an aggregate query.

    Args:
        func:
        query:
        user:

    Returns:

    """
    if user.is_superuser:
        return func(query, user)

    # update the query
    query = _update_can_read_aggregate_query(query, user)
    # get list of data
    data = func(query, user)

    return data


def _update_can_read_query(query, user):
    """ Update query with access control parameters.

    Args:
        query:
        user:

    Returns:

    """

    accessible_workspaces = _get_read_accessible_workspaces_by_user(user)
    # update query with workspace criteria
    query = add_access_criteria(query, accessible_workspaces, user)
    return query


def _update_can_read_aggregate_query(query, user):
    """ Update query with access control parameters.

    Args:
        query:
        user:

    Returns:

    """

    accessible_workspaces = _get_read_accessible_workspaces_by_user(user)
    # update query with workspace criteria
    query = add_aggregate_access_criteria(query, accessible_workspaces, user)
    return query


def _get_read_accessible_workspaces_by_user(user):
    """ Get read accessible workspaces by user.

    Args:
        user:

    Returns:

    """
    if not CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT and user.is_anonymous:
        accessible_workspaces = []
    else:
        # workspace case
        # list accessible workspaces
        accessible_workspaces = [workspace.id for workspace in
                                 workspace_api.get_all_workspaces_with_read_access_by_user(user)]

    return accessible_workspaces


def can_write_data_workspace(func, data, workspace, user):
    """ Can user write data in workspace.

    Args:
        func:
        data:
        workspace:
        user:

    Returns:

    """
    return can_write_in_workspace(func, data, workspace, user, rights.publish_data)



""" Unit Test Data
"""
from collections import OrderedDict
from unittest.case import TestCase

from mock import patch

import core_main_app.components.data.api as data_api
from core_main_app.commons import exceptions
from core_main_app.components.data.models import Data
from core_main_app.components.template.models import Template
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.utils.tests_tools.MockUser import create_mock_user


class TestDataGetById(TestCase):

    @patch.object(Data, 'get_by_id')
    def test_data_get_by_id_raises_api_error_if_not_found(self, mock_get):
        # Arrange
        mock_get.side_effect = exceptions.DoesNotExist('')
        mock_user = _create_user('1')
        # Act # Assert
        with self.assertRaises(exceptions.DoesNotExist):
            data_api.get_by_id(1, mock_user)

    @patch.object(Data, 'get_by_id')
    def test_data_get_by_id_return_data_if_found(self, mock_get):
        # Arrange
        mock_data = Data(template=_get_template(), user_id='1', dict_content=OrderedDict(),
                         title='title')
        mock_get.return_value = mock_data
        mock_user = _create_user('1')
        # Act
        result = data_api.get_by_id(1, mock_user)
        # Assert
        self.assertIsInstance(result, Data)


class TestDataGetAllExceptUser(TestCase):

    @patch("core_main_app.components.workspace.api.get_all_workspaces_with_read_access_by_user")
    @patch.object(Data, 'get_all_except_user_id')
    def test_data_get_all_except_user_return_collection_of_data_where_user_is_not_owner(self, mock_list_except_user_id,
                                                                                        get_all_workspaces_with_read_access_by_user):
        # Arrange
        user_id = '2'
        mock_data_1 = _create_data(_get_template(), user_id='3', title='title_1', content="")
        mock_data_2 = _create_data(_get_template(), user_id='1', title='title_2', content="")
        mock_list_except_user_id.return_value = [mock_data_1, mock_data_2]
        mock_user = _create_user('2')
        get_all_workspaces_with_read_access_by_user.return_value = []
        # Act
        with self.assertRaises(AccessControlError):
            data_api.get_all_except_user(mock_user)


class TestDataUpsert(TestCase):

    @patch.object(Data, 'convert_to_file')
    @patch.object(data_api, 'check_xml_file_is_valid')
    @patch.object(Data, 'save')
    def test_data_upsert_return_data_with_new_title_if_is_called_with_only_title_modified(self, mock_save,
                                                                                          mock_check,
                                                                                          mock_convert_file):
        # Arrange
        data = _create_data(_get_template(), user_id='2', title='new_title', content='<tag></tag>')
        mock_save.return_value = data
        mock_check.return_value = None
        mock_convert_file.return_value = None
        mock_user = _create_user('2')
        # Act
        result = data_api.upsert(data, mock_user)
        # Assert
        self.assertEqual('new_title', result.title)

    @patch.object(Data, 'convert_to_file')
    @patch.object(data_api, 'check_xml_file_is_valid')
    @patch.object(Data, 'save')
    def test_data_upsert_return_data_with_new_user_id_if_is_called_with_only_user_id_modified(self, mock_save,
                                                                                              mock_check,
                                                                                              mock_convert_file):
        # Arrange
        data = _create_data(_get_template(), user_id='3', title='new_title', content='<tag></tag>')
        mock_save.return_value = data
        mock_check.return_value = None
        mock_convert_file.return_value = None
        mock_user = _create_user('3')
        # Act
        result = data_api.upsert(data, mock_user)
        # Assert
        self.assertEqual('3', result.user_id)

    @patch.object(Data, 'convert_to_file')
    @patch.object(data_api, 'check_xml_file_is_valid')
    @patch.object(Data, 'save')
    def test_data_upsert_return_data_with_new_xml_if_is_called_with_only_xml_modified(self, mock_save,
                                                                                      mock_check, mock_convert_file):
        # Arrange
        xml = '<new_tag></new_tag>'
        data = _create_data(_get_template(), user_id='3', title='title', content=xml)
        mock_save.return_value = data
        mock_check.return_value = None
        mock_convert_file.return_value = None
        mock_user = _create_user('3')
        # Act
        result = data_api.upsert(data, mock_user)
        # Assert
        self.assertEqual(xml, result.xml_content)

    @patch.object(Data, 'convert_to_file')
    @patch.object(data_api, 'check_xml_file_is_valid')
    @patch.object(Data, 'save')
    def test_data_upsert_return_data_with_last_modification_date(self, mock_save,
                                                                 mock_check, mock_convert_file):
        # Arrange
        data = _create_data(_get_template(), user_id='3', title='title', content='<tag></tag>')
        mock_save.return_value = data
        mock_check.return_value = None
        mock_convert_file.return_value = None
        mock_user = _create_user('3')
        # Act
        result = data_api.upsert(data, mock_user)
        # Assert
        self.assertIsNotNone(result.last_modification_date)

    def test_data_upsert_raises_xml_error_if_failed_during_xml_validation(self):
        # Arrange
        data = _create_data(None, user_id='3', title='title', content='')
        mock_user = _create_user('3')
        # Act # Assert
        with self.assertRaises(exceptions.XMLError):
            data_api.upsert(data, mock_user)

    def test_data_upsert_raises_xsd_error_if_failed_during_xsd_validation(self):
        # Arrange
        template = _get_template()
        template.content += "<"
        data = _create_data(template, user_id='3', title='title', content='<new_tag></new_tag>')
        mock_user = _create_user('3')
        # Act # Assert
        with self.assertRaises(exceptions.XSDError):
            data_api.upsert(data, mock_user)

    def test_data_upsert_raises_xml_error_if_failed_during_validation(self):
        # Arrange
        template = _get_template()
        data = _create_data(template, user_id='3', title='title', content='<new_tag></new_tag>')
        mock_user = _create_user('3')
        # Act # Assert
        with self.assertRaises(exceptions.XMLError):
            data_api.upsert(data, mock_user)


class TestDataCheckXmlFileIsValid(TestCase):

    def test_data_check_xml_file_is_valid_raises_xml_error_if_failed_during_xml_validation(self):
        # Arrange
        data = _create_data(None, user_id='3', title='title', content='')
        # Act # Assert
        with self.assertRaises(exceptions.XMLError):
            data_api.check_xml_file_is_valid(data)

    def test_data_check_xml_file_is_valid_raises_xsd_error_if_failed_during_xsd_validation(self):
        # Arrange
        template = _get_template()
        template.content += "<"
        data = _create_data(template, user_id='3', title='title', content='<new_tag></new_tag>')
        # Act # Assert
        with self.assertRaises(exceptions.XSDError):
            data_api.check_xml_file_is_valid(data)

    def test_data_check_xml_file_is_valid_raises_xml_error_if_failed_during_validation(self):
        # Arrange
        template = _get_template()
        data = _create_data(template, user_id='3', title='title', content='<new_tag></new_tag>')
        # Act # Assert
        with self.assertRaises(exceptions.XMLError):
            data_api.check_xml_file_is_valid(data)

    def test_data_check_xml_data_valid_return_true_if_validation_success(self):
        # Arrange
        template = _get_template()
        data = _create_data(template, user_id='3', title='title', content='<tag>toto</tag>')
        # Act
        result = data_api.check_xml_file_is_valid(data)
        # Assert
        self.assertEqual(result, True)


def _get_template():
    template = Template()
    template.id_field = 1
    xsd = '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">' \
          '<xs:element name="tag"></xs:element></xs:schema>'
    template.content = xsd
    return template


def _create_data(template, user_id, title, content):
    data = Data(template=template,
                user_id=user_id,
                title=title)
    data.xml_content = content
    return data


def _create_user(user_id):
    return create_mock_user(user_id)



"""Serializers used throughout the Rest API
"""
from rest_framework_mongoengine.serializers import DocumentSerializer

from core_main_app.components.template.api import init_template_with_dependencies
from core_main_app.components.template.models import Template
from core_main_app.components.template_version_manager import api as template_version_manager_api
from core_main_app.components.template_version_manager.models import TemplateVersionManager
from core_main_app.rest.template.serializers import TemplateSerializer
from core_main_app.rest.template_version_manager.utils import load_dependencies


class TemplateVersionManagerSerializer(DocumentSerializer):
    """
        Template Version Manager serializer
    """
    class Meta(object):
        model = TemplateVersionManager
        fields = "__all__"
        read_only_fields = ['id',
                            'user',
                            'versions',
                            'current',
                            'is_disabled',
                            'disabled_versions']

    def create(self, validated_data):
        """ Create.

        Args:
            validated_data:

        Returns:

        """
        return TemplateVersionManager(**validated_data)


class CreateTemplateSerializer(TemplateSerializer):
    """
        Template Version Manager serializer
    """
    def create(self, validated_data):
        """
        Create and return a new `Template` instance, given the validated data.
        """
        template_object = Template(filename=validated_data['filename'],
                                   content=validated_data['content'])
        template_version_manager_object = validated_data['template_version_manager']

        # load dependencies
        dependencies_dict = load_dependencies(validated_data)

        # Update the content of the template with dependencies
        init_template_with_dependencies(template_object, dependencies_dict)

        # Create the template and its template version manager
        template_version_manager_api.insert(template_version_manager_object, template_object)

        return template_object

    def update(self, instance, validated_data):
        raise NotImplementedError("Template Version Manager should only be updated using specialized APIs.")



""" REST views for the data API
"""
import json

from django.http import Http404
from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated, IsAuthenticatedOrReadOnly, IsAdminUser
from rest_framework.response import Response
from rest_framework.views import APIView

from core_main_app.access_control.api import check_can_write
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons import exceptions
from core_main_app.components.data import api as data_api
from core_main_app.components.user import api as user_api
from core_main_app.components.workspace import api as workspace_api
from core_main_app.rest.data.abstract_views import AbstractExecuteLocalQueryView
from core_main_app.rest.data.serializers import DataSerializer, DataWithTemplateInfoSerializer
from core_main_app.utils.boolean import to_bool
from core_main_app.utils.databases.pymongo_database import get_full_text_query
from core_main_app.utils.file import get_file_http_response
from core_main_app.utils.pagination.rest_framework_paginator.pagination import StandardResultsSetPagination


class DataList(APIView):
    """ List all user Data, or create a new one.
    """
    permission_classes = (IsAuthenticated,)
    serializer = DataSerializer

    def get(self, request):
        """ Get all user Data

        Url Parameters:

            template: template_id
            title: document_title

        Examples:

            ../data/
            ../data?workspace=[workspace_id]
            ../data?template=[template_id]
            ../data?title=[document_title]
            ../data?template=[template_id]&title=[document_title]

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of data
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            data_object_list = data_api.get_all_by_user(request.user)

            # Apply filters
            workspace = self.request.query_params.get('workspace', None)
            if workspace is not None:
                data_object_list = data_object_list.filter(workspace=workspace)

            template = self.request.query_params.get('template', None)
            if template is not None:
                data_object_list = data_object_list.filter(template=template)

            title = self.request.query_params.get('title', None)
            if title is not None:
                data_object_list = data_object_list.filter(title=title)

            # Serialize object
            data_serializer = self.serializer(data_object_list, many=True)

            # Return response
            return Response(data_serializer.data, status=status.HTTP_200_OK)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def post(self, request):
        """ Create a Data

        Parameters:

            {
                "title": "document_title",
                "template": "template_id",
                "workspace": "workspace_id",
                "xml_content": "document_content"
            }

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Created data
            - code: 400
              content: Validation error
            - code: 404
              content: Template was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializer
            data_serializer = self.serializer(data=request.data)

            # Validate data
            data_serializer.is_valid(True)
            # Save data
            data_serializer.save(user=request.user)

            # Return the serialized data
            return Response(data_serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except exceptions.DoesNotExist:
            content = {'message': 'Template not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class DataDetail(APIView):
    """ Retrieve, update or delete a Data
    """
    permission_classes = (IsAuthenticatedOrReadOnly,)
    serializer = DataSerializer

    def get_object(self, request, pk):
        """ Get data from db

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            Data
        """
        try:
            return data_api.get_by_id(pk, request.user)
        except exceptions.DoesNotExist:
            raise Http404

    def get(self, request, pk):
        """ Retrieve a data

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: Data
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            data_object = self.get_object(request, pk)

            # Serialize object
            serializer = self.serializer(data_object)

            # Return response
            return Response(serializer.data)
        except Http404:
            content = {'message': 'Data not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def delete(self, request, pk):
        """ Delete a Data

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 204
              content: Deletion succeed
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            data_object = self.get_object(request, pk)

            # delete object
            data_api.delete(data_object, request.user)

            # Return response
            return Response(status=status.HTTP_204_NO_CONTENT)
        except Http404:
            content = {'message': 'Data not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def patch(self, request, pk):
        """ Update a Data

        Parameters:

            {
                "title": "new_title",
                "xml_content": "new_xml_content"
            }

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: Updated data
            - code: 400
              content: Validation error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            data_object = self.get_object(request, pk)

            # Build serializer
            data_serializer = self.serializer(instance=data_object,
                                              data=request.data,
                                              partial=True)

            # Validate data
            data_serializer.is_valid(True)
            # Save data
            data_serializer.save(user=request.user)

            return Response(data_serializer.data, status=status.HTTP_200_OK)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Http404:
            content = {'message': 'Data not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class DataChangeOwner(APIView):
    """ Change the Owner of a data
    """
    permission_classes = (IsAdminUser,)

    def get_object(self, request, pk):
        """ Get data from db

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            Data
        """
        try:
            return data_api.get_by_id(pk, request.user)
        except exceptions.DoesNotExist:
            raise Http404

    def get_user(self, user_id):
        """ Retrieve a User

        Args:

            user_id: ObjectId

        Returns:

            - code: 404
              content: Object was not found
        """
        try:
            return user_api.get_user_by_id(user_id)
        except exceptions.DoesNotExist:
            raise Http404

    def patch(self, request, pk, user_id):
        """ Change the Owner of a data

        Args:

            request: HTTP request
            pk: ObjectId
            user_id: ObjectId

        Returns:

            - code: 200
              content: None
            - code: 403
              content: Authentication error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # get object
            data_object = self.get_object(request, pk)
            user_object = self.get_user(user_id)

            # change owner
            data_api.change_owner(data_object, user_object, request.user)
            return Response({}, status=status.HTTP_200_OK)
        except Http404:
            content = {'message': 'Data or user not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except AccessControlError as ace:
            content = {'message': str(ace)}
            return Response(content, status=status.HTTP_403_FORBIDDEN)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class DataDownload(APIView):
    """ Download XML file in data
    """

    def get_object(self, request, pk):
        """ Get Data from db

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            Data
        """
        try:
            return data_api.get_by_id(pk, request.user)
        except exceptions.DoesNotExist:
            raise Http404

    def get(self, request, pk):
        """ Download the XML file from a data

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: XML file
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            data_object = self.get_object(request, pk)

            return get_file_http_response(data_object.xml_content, data_object.title, 'text/xml', 'xml')
        except Http404:
            content = {'message': 'Data not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


# FIXME: Should use in the future an serializer with dynamic fields (init depth with parameter for example)
# FIXME: Should avoid the duplicated code with get_by_id
@api_view(['GET'])
def get_by_id_with_template_info(request):
    """ Retrieve a Data with template information

    Examples:

        ../data/get-full?id=[data_id]

    Args:

        request: HTTP request

    Returns:

        - code: 200
          content: Data
        - code: 400
          content: Validation error
        - code: 404
          content: Object was not found
        - code: 500
          content: Internal server error
    """
    try:
        # Get parameters
        data_id = request.query_params.get('id', None)

        # Check parameters
        if data_id is None:
            content = {'message': 'Expected parameters not provided.'}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)

        # Get object
        data_object = data_api.get_by_id(data_id, request.user)

        # Serialize object
        return_value = DataWithTemplateInfoSerializer(data_object)

        # Return response
        return Response(return_value.data, status=status.HTTP_200_OK)
    except exceptions.DoesNotExist as e:
        content = {'message': 'No data found with the given id.'}
        return Response(content, status=status.HTTP_404_NOT_FOUND)
    except exceptions.ModelError:
        content = {'message': 'Invalid input.'}
        return Response(content, status=status.HTTP_400_BAD_REQUEST)
    except Exception:
        content = {'message': 'An unexpected error occurred.'}
        return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class ExecuteLocalQueryView(AbstractExecuteLocalQueryView):
    serializer = DataSerializer

    def post(self, request):
        """ Execute a query

        Url Parameters:

            page: page_number

        Parameters:

            # get all results (paginated)
            {"query": "{}"}
            # get all results
            {"query": "{}", "all": "true"}
            # get all results filtered by title
            {"query": "{}", "title": "title_string"}
            # get all results filtered by workspaces
            {"query": "{}", "workspaces": "[{\\"id\\":\\"[workspace_id]\\"}]"}
            # get all results filtered by private workspace
            {"query": "{}", "workspaces": "[{\\"id\\":\\"None\\"}]"}
            # get all results filtered by templates
            {"query": "{}", "templates": "[{\\"id\\":\\"[template_id]\\"}]"}
            # get all results that verify a given criteria
            {"query": "{\\"root.element.value\\": 2}"}
            # get results using multiple options
            {"query": "{\\"root.element.value\\": 2}", "workspaces": "[{\\"id\\":\\"workspace_id\\"}]", "all": "true"}
            {"query": "{\\"root.element.value\\": 2}", "templates": "[{\\"id\\":\\"template_id\\"}]", "all": "true"}
            {"query": "{\\"root.element.value\\": 2}", "templates": "[{\\"id\\":\\"template_id\\"}]",
            "workspaces": "[{\\"id\\":\\"[workspace_id]\\"}]","all": "true"}

        Warning:

            Need to backslash double quotes in JSON payload

        Examples:

            ../data/query/
            ../data/query/?page=2

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of data
            - code: 400
              content: Bad request
            - code: 500
              content: Internal server error
        """
        return super(ExecuteLocalQueryView, self).post(request)

    def build_response(self, data_list):
        """ Build the response.

        Args:

            data_list: List of data

        Returns:

            The response paginated
        """
        if 'all' in self.request.data and to_bool(self.request.data['all']):
            # Serialize data list
            data_serializer = self.serializer(data_list, many=True)
            # Return response
            return Response(data_serializer.data)
        else:
            # Get paginator
            paginator = StandardResultsSetPagination()

            # Get requested page from list of results
            page = paginator.paginate_queryset(data_list, self.request)

            # Serialize page
            data_serializer = self.serializer(page, many=True)

            # Return paginated response
            return paginator.get_paginated_response(data_serializer.data)


class ExecuteLocalKeywordQueryView(ExecuteLocalQueryView):
    def build_query(self, query, workspaces, templates, options, title=None):
        """ Build the raw query
        Prepare the query for a keyword search

        Args:

            query: ObjectId
            workspaces: ObjectId
            templates: ObjectId
            options: Query options
            title: title filter
            
        Returns:

            The raw query
        """
        # build query builder
        query = json.dumps(get_full_text_query(query))
        return super(ExecuteLocalKeywordQueryView, self).build_query(str(query), templates, options, workspaces, title)


class DataAssign(APIView):
    """ Assign a Data to a Workspace.
    """
    permission_classes = (IsAuthenticated,)

    def get_object(self, request, pk):
        """ Get data from db

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            Data
        """
        try:
            return data_api.get_by_id(pk, request.user)
        except exceptions.DoesNotExist:
            raise Http404

    def get_workspace(self, workspace_id):
        """ Retrieve a Workspace

        Args:

            workspace_id: ObjectId

        Returns:

            - code: 404
              content: Object was not found
        """
        try:
            return workspace_api.get_by_id(workspace_id)
        except exceptions.DoesNotExist:
            raise Http404

    def patch(self, request, pk, workspace_id):
        """ Assign Data to a Workspace

        Args:

            request: HTTP request
            pk: ObjectId
            workspace_id: ObjectId

        Returns:

            - code: 200
              content: None
            - code: 403
              content: Authentication error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            data_object = self.get_object(request, pk)
            workspace_object = self.get_workspace(workspace_id)

            # Assign data to workspace
            data_api.assign(data_object, workspace_object, request.user)
            return Response({}, status=status.HTTP_200_OK)
        except Http404:
            content = {'message': 'Data or workspace not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except AccessControlError as ace:
            content = {'message': str(ace)}
            return Response(content, status=status.HTTP_403_FORBIDDEN)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class DataListByWorkspace(APIView):
    """ List all Data by workspace.
    """
    permission_classes = (IsAuthenticated,)
    serializer = DataSerializer

    def get(self, request, workspace_id):
        """ Get all workspace Data

        Examples:

            ../workspace/id/data


        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of data
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            data_object_list = data_api.get_all_by_workspace(workspace_id, request.user)

            # Serialize object
            data_serializer = self.serializer(data_object_list, many=True)

            # Return response
            return Response(data_serializer.data, status=status.HTTP_200_OK)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class DataPermissions(APIView):
    """
    Get the permissions of the data according to the client user
    """

    def get_object(self, request, pk):
        """ Get data from db

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            Data
        """
        try:
            return data_api.get_by_id(pk, request.user)
        except exceptions.DoesNotExist:
            raise Http404

    def get(self, request):
        """ Give the user permissions for a list of data ids

        Parameters:

            [
                "data_id1"
                "data_id2"
                "data_id3"
                ...
            ]

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: JSON Array [ <data_id>: <boolean> ]
            - code: 400
              content: Validation error
            - code: 404
              content: Template was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializer
            data_ids = json.loads(request.query_params['ids'])
            results = {}

            for id in data_ids:
                results[id] = (self.can_write_data(request, id))

            return Response(results, status.HTTP_200_OK)

        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Http404:
            content = {'message': 'Data not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def can_write_data(self, request, id):
        """ Get the data permissions of a data

        Args:

            request: http request
            id: data id

        Returns:

            - Boolean
            - raise: Http404
              content: Data was not found
            - raise: Exception
              content: Unknown error
        """
        try:
            # Get object
            data_object = self.get_object(request, id)

            if not request.user.is_superuser:
                check_can_write(data_object, request.user)
            return True
        except AccessControlError as ace:
            return False
        except Exception as e:
            raise e



""" REST abstract views for the template version manager API
"""
from abc import ABCMeta, abstractmethod

from django.http import Http404
from django.utils.decorators import method_decorator
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView

from core_main_app.commons import exceptions as exceptions
from core_main_app.commons.exceptions import NotUniqueError, ApiError, XSDError
from core_main_app.components.template import api as template_api
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.rest.template_version_manager.serializers import TemplateVersionManagerSerializer, \
    CreateTemplateSerializer
from core_main_app.rest.template_version_manager.utils import can_user_modify_template_version_manager
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.utils.boolean import to_bool
from core_main_app.utils.decorators import api_staff_member_required


class AbstractTemplateVersionManagerList(APIView, metaclass=ABCMeta):
    """ List template version managers
    """
    serializer = TemplateVersionManagerSerializer

    @abstractmethod
    def get_template_version_managers(self):
        """ Return template version managers
        """
        raise NotImplementedError("get_template_version_managers method is not implemented.")

    def get(self, request):
        """ Get template version managers

        Url Parameters:

            template: template_id
            title: document_title

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of template version manager
            - code: 500
              content: Internal server error
        """
        try:
            # Get objects
            object_list = self.get_template_version_managers()

            # Apply filters
            title = self.request.query_params.get('title', None)
            if title is not None:
                object_list = object_list.filter(title=title)

            is_disabled = self.request.query_params.get('is_disabled', None)
            if is_disabled is not None:
                object_list = object_list.filter(is_disabled=to_bool(is_disabled))

            # Serialize object
            serializer = self.serializer(object_list, many=True)
            return Response(serializer.data, status=status.HTTP_200_OK)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class AbstractStatusTemplateVersion(APIView, metaclass=ABCMeta):
    """ Set template version status
    """

    def get_object(self, pk):
        """ Get template from db

        Args:

            pk: ObjectId

        Returns:

            Template
        """
        try:
            template_object = template_api.get(pk)
            template_version_manager_object = version_manager_api.get_from_version(template_object)
            can_user_modify_template_version_manager(template_version_manager_object, self.request.user)
            return template_object
        except exceptions.DoesNotExist:
            raise Http404

    @abstractmethod
    def status_update(self, template_object):
        """ Perform an update of the object status
        """
        raise NotImplementedError("status_update method is not implemented.")

    @method_decorator(api_staff_member_required())
    def patch(self, request, pk):
        """ Set status

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: None
            - code: 400
              content: Validation error / bad request
            - code: 403
              content: Authentication error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            template_object = self.get_object(pk)

            # Set current template
            self.status_update(template_object)

            return Response(status=status.HTTP_200_OK)
        except Http404:
            content = {'message': 'Template not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except AccessControlError as access_error:
            content = {'message': str(access_error)}
            return Response(content, status=status.HTTP_403_FORBIDDEN)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except ApiError as api_error:
            content = {'message': str(api_error)}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class AbstractTemplateVersionManagerDetail(APIView, metaclass=ABCMeta):
    """ Template Version Manager Detail
    """

    def get_object(self, pk):
        """ Get template version manager from db

        Args:

            pk: ObjectId

        Returns:

            TemplateVersionManager
        """
        try:
            template_version_manager_object = version_manager_api.get(pk)
            can_user_modify_template_version_manager(template_version_manager_object, self.request.user)
            return template_version_manager_object
        except exceptions.DoesNotExist:
            raise Http404


class AbstractStatusTemplateVersionManager(AbstractTemplateVersionManagerDetail, metaclass=ABCMeta):
    """ Set template version manager status
    """

    @abstractmethod
    def status_update(self, template_version_manager_object):
        """ Perform an update of the object status.
        """
        raise NotImplementedError("status_update method is not implemented.")

    def patch(self, request, pk):
        """ Set status

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: None
            - code: 400
              content: Validation error
            - code: 403
              content: Authentication error
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            template_version_manager_object = self.get_object(pk)

            # Set current template
            self.status_update(template_version_manager_object)

            return Response(status=status.HTTP_200_OK)
        except Http404:
            content = {'message': 'Template Version Manager not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except AccessControlError as access_error:
            content = {'message': str(access_error)}
            return Response(content, status=status.HTTP_403_FORBIDDEN)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class AbstractTemplateList(APIView, metaclass=ABCMeta):
    """ Create a template
    """
    serializer = TemplateVersionManagerSerializer
    create_serializer = CreateTemplateSerializer

    def post(self, request):
        """ Create a template

        Parameters:

            {
                "title": "title",
                "filename": "filename",
                "content": "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'><xs:element name='root'/></xs:schema>"
            }

        Note: 

            "dependencies_dict": json.dumps({"schemaLocation1": "id1" ,"schemaLocation2":"id2"})

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Created template
            - code: 400
              content: Validation error / not unique / XSD error
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializers
            template_serializer = self.create_serializer(data=request.data)
            template_version_manager_serializer = self.serializer(data=request.data)

            # Validate data
            template_serializer.is_valid(True)
            template_version_manager_serializer.is_valid(True)

            # Save data
            template_version_manager_object = template_version_manager_serializer.save(user=self.get_user())
            template_serializer.save(template_version_manager=template_version_manager_object)

            return Response(template_serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except NotUniqueError:
            content = {'message': "A template with the same title already exists."}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except XSDError as xsd_error:
            content = {'message': "XSD Error: " + str(xsd_error)}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @abstractmethod
    def get_user(self):
        """ Retrieve a user
        """
        raise NotImplementedError("get_user method is not implemented.")



""" Mongo engine database tools
"""
from mongoengine import connect
from mongoengine.connection import disconnect


class Database(object):
    """ Represent a Database.
    """

    def __init__(self, host, name):
        """ Constructor of the database Object

        Args:
            host: host of the database (like an uri)
            name: name of the database
        """
        self.database_host = host
        self.database_name = name
        self.database = None

    def connect(self):
        """ Open a connection to the database.

        Returns:
            the database connection created
        """
        self.database = connect(self.database_name, host=self.database_host)
        return self.database

    def disconnect(self):
        """ Close the connection.
        """
        disconnect(self.database)

    def clean_database(self):
        """ Clear all collections of the database.

        Returns:

        """
        # Clear the database for the next test
        if self.database[self.database_name] is not None:
            # Clear all collections
            for collection in self.database[self.database_name].list_collection_names():
                if collection != 'system.indexes':
                    # WARNING: Do not drop the collection but clear it. Drop collection with mongomock is not well
                    # supported. Please see https://github.com/mongomock/mongomock/issues/238
                    self.database[self.database_name].get_collection(collection).delete_many({})


def init_text_index(document_object):
    """ Create index for full text search.
    """
    collection = document_object._get_collection()
    collection.create_index([('$**', "text")], default_language="en", language_override="en")



""" Ajax API
"""
import json

from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest
from django.template import loader
from django.utils.html import escape
from django.views.generic import View

from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import DoesNotExist, NotUniqueError, ModelError
from core_main_app.components.group import api as group_api
from core_main_app.components.user import api as user_api
from core_main_app.components.workspace import api as workspace_api
from core_main_app.utils import group as group_utils
from core_main_app.views.user.forms import ChangeWorkspaceForm, UserRightForm, GroupRightForm

GROUP = "group"
USER = "user"

ACTION_READ = "action_read"
ACTION_WRITE = "action_write"


@login_required
def set_public_workspace(request):
    """ Set a workspace public.

    Args:
        request:

    Returns:
    """
    workspace_id_list = request.POST.getlist('workspace_id[]', [])
    try:
        list_workspace = workspace_api.get_by_id_list(workspace_id_list)
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    try:
        for workspace in list_workspace:
            workspace_api.set_workspace_public(workspace, request.user)
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except:
        return HttpResponseBadRequest("Something wrong happened.")

    return HttpResponse(json.dumps({}), content_type='application/javascript')


@login_required
def set_private_workspace(request):
    """ Set a workspace private.

    Args:
        request:

    Returns:
    """
    workspace_id_list = request.POST.getlist('workspace_id[]', [])
    try:
        list_workspace = workspace_api.get_by_id_list(workspace_id_list)
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    try:
        for workspace in list_workspace:
            workspace_api.set_workspace_private(workspace, request.user)
    except:
        return HttpResponseBadRequest("Something wrong happened.")

    return HttpResponse(json.dumps({}), content_type='application/javascript')


class LoadFormChangeWorkspace(View):
    """ Load the form to list the workspaces. """

    show_global_workspace = True

    def post(self, request, *args, **kwargs):

        is_administration = request.POST.get('administration', False) == "True"

        try:
            form = ChangeWorkspaceForm(request.user, list(), is_administration, self.show_global_workspace)
        except DoesNotExist as dne:
            return HttpResponseBadRequest(escape(str(dne)))
        except Exception as e:
            return HttpResponseBadRequest("Something wrong happened.")

        context = {
            "assign_workspace_form": form
        }

        return HttpResponse(json.dumps({'form': loader.render_to_string(
            "core_main_app/user/workspaces/list/modals/assign_workspace_form.html", context)}),
            'application/javascript')


@login_required
def create_workspace(request):
    """ Create a workspace.

    Args:
        request

    Returns:
    """
    name_workspace = request.POST.get('name_workspace', None)
    try:
        workspace_api.create_and_save(name_workspace, request.user.id)
    except NotUniqueError:
        return HttpResponseBadRequest("You already have a workspace called "
                                      + name_workspace +
                                      ". Please change the name and try again.")
    except exceptions.ModelError:
        return HttpResponseBadRequest("The given title does not meet the requirements for a title. "
                                      "Please provide another title.")
    except Exception as e:
        return HttpResponseBadRequest("A problem occurred while creating the workspace.")
    return HttpResponse(json.dumps({}), content_type='application/javascript')


@login_required
def load_add_user_form(request):
    """ Load the form to list the users with no access to the workspace.

    Args:
        request:

    Returns:
    """
    workspace_id = request.POST.get('workspace_id', None)
    try:
        workspace = workspace_api.get_by_id(str(workspace_id))
    except exceptions.ModelError:
        return HttpResponseBadRequest('Invalid input.')
    except Exception:
        return HttpResponseBadRequest('An unexpected error occurred.')

    try:
        # We retrieve all users with no access
        users_with_no_access = list(workspace_api.get_list_user_with_no_access_workspace(workspace, request.user))

        # We remove the owner of the workspace
        if len(users_with_no_access) > 0:
            users_with_no_access.remove(user_api.get_user_by_id(workspace.owner))

        if len(users_with_no_access) == 0:
            return HttpResponseBadRequest("There is no users that can be added.")

        form = UserRightForm(users_with_no_access)
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except:
        return HttpResponseBadRequest("Something wrong happened.")

    context = {
        "add_user_form": form
    }

    return HttpResponse(json.dumps({'form': loader.render_to_string(
        "core_main_app/user/workspaces/list/modals/add_user_form.html", context)}),
        'application/javascript')


@login_required
def add_user_right_to_workspace(request):
    """ Add rights to user for the workspace.

    Args:
        request

    Returns
    """
    workspace_id = request.POST.get('workspace_id', None)
    users_ids = request.POST.getlist('users_id[]', [])
    is_read_checked = request.POST.get('read', None) == 'true'
    is_write_checked = request.POST.get('write', None) == 'true'

    if len(users_ids) == 0:
        return HttpResponseBadRequest("You need to select at least one user.")
    if not is_read_checked and not is_write_checked:
        return HttpResponseBadRequest("You need to select at least one permission (read and/or write).")

    try:
        workspace = workspace_api.get_by_id(str(workspace_id))
        for user in user_api.get_all_users_by_list_id(users_ids):
            if is_read_checked:
                workspace_api.add_user_read_access_to_workspace(workspace, user, request.user)
            if is_write_checked:
                workspace_api.add_user_write_access_to_workspace(workspace, user, request.user)
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception as exc:
        return HttpResponseBadRequest('Something wrong happened.')

    return HttpResponse(json.dumps({}), content_type='application/javascript')


@login_required
def switch_right(request):
    """ Switch user's right for the workspace.

    Args:
        request

    Returns
    """

    workspace_id = request.POST.get('workspace_id', None)
    object_id = request.POST.get('object_id', None)
    group_or_user = request.POST.get('group_or_user', None)
    action = request.POST.get('action', None)
    value = request.POST.get('value', None) == 'true'

    try:
        workspace = workspace_api.get_by_id(str(workspace_id))

        if group_or_user == USER:
            _switch_user_right(object_id, action, value, workspace, request.user)
        if group_or_user == GROUP:
            _switch_group_right(object_id, action, value, workspace, request.user)

    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception as exc:
        return HttpResponseBadRequest('Something wrong happened.')

    return HttpResponse(json.dumps({}), content_type='application/javascript')


def _switch_user_right(user_id, action, value, workspace, request_user):
    """ Change the user rights to the workspace.

    Args:
        user_id:
        action:
        value:
        workspace:
        request_user:

    Returns:
    """
    user = user_api.get_user_by_id(user_id)

    if action == ACTION_READ:
        if value:
            workspace_api.add_user_read_access_to_workspace(workspace, user, request_user)
        else:
            workspace_api.remove_user_read_access_to_workspace(workspace, user, request_user)
    elif action == ACTION_WRITE:
        if value:
            workspace_api.add_user_write_access_to_workspace(workspace, user, request_user)
        else:
            workspace_api.remove_user_write_access_to_workspace(workspace, user, request_user)


def _switch_group_right(group_id, action, value, workspace, request_user):
    """ Change the group rights to the workspace.

    Args:
        group_id:
        action:
        value:
        workspace:
        request_user:

    Returns:
    """
    group = group_api.get_group_by_id(group_id)

    if action == ACTION_READ:
        if value:
            workspace_api.add_group_read_access_to_workspace(workspace, group, request_user)
        else:
            workspace_api.remove_group_read_access_to_workspace(workspace, group, request_user)
    elif action == ACTION_WRITE:
        if value:
            workspace_api.add_group_write_access_to_workspace(workspace, group, request_user)
        else:
            workspace_api.remove_group_write_access_to_workspace(workspace, group, request_user)


@login_required
def remove_user_or_group_rights(request):
    """ Remove user's right for the workspace.

    Args:
        request

    Returns
    """

    workspace_id = request.POST.get('workspace_id', None)
    object_id = request.POST.get('object_id', None)
    group_or_user = request.POST.get('group_or_user', None)

    try:
        workspace = workspace_api.get_by_id(str(workspace_id))

        if group_or_user == USER:
            _remove_user_rights(object_id, workspace, request.user)
        if group_or_user == GROUP:
            _remove_group_rights(object_id, workspace, request.user)

    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except ModelError:
        return HttpResponseBadRequest('Invalid input.')
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception:
        return HttpResponseBadRequest('Something wrong happened.')

    return HttpResponse(json.dumps({}), content_type='application/javascript')


def _remove_user_rights(object_id, workspace, request_user):
    """ Remove all user rights on the workspace.

    Args:
        object_id:
        workspace:
        request_user:

    Returns:
    """
    user = user_api.get_user_by_id(object_id)
    workspace_api.remove_user_read_access_to_workspace(workspace, user, request_user)
    workspace_api.remove_user_write_access_to_workspace(workspace, user, request_user)


def _remove_group_rights(object_id, workspace, request_user):
    """ Remove all group rights on the workspace.

    Args:
        object_id:
        workspace:
        request_user:

    Returns:
    """
    group = group_api.get_group_by_id(object_id)
    workspace_api.remove_group_read_access_to_workspace(workspace, group, request_user)
    workspace_api.remove_group_write_access_to_workspace(workspace, group, request_user)


@login_required
def load_add_group_form(request):
    """ Load the form to list the groups with no access to the workspace.

    Args:
        request:

    Returns:
    """
    workspace_id = request.POST.get('workspace_id', None)
    try:
        workspace = workspace_api.get_by_id(str(workspace_id))
    except exceptions.ModelError:
        return HttpResponseBadRequest('Invalid input.')
    except Exception:
        return HttpResponseBadRequest('An unexpected error occurred.')

    try:
        # We retrieve all groups with no access
        groups_with_no_access = list(workspace_api.get_list_group_with_no_access_workspace(workspace, request.user))

        if len(groups_with_no_access) > 0:
            group_utils.remove_list_object_from_list(groups_with_no_access,
                                                     [group_api.get_anonymous_group(), group_api.get_default_group()])
        if len(groups_with_no_access) == 0:
            return HttpResponseBadRequest("There is no groups that can be added.")

        form = GroupRightForm(groups_with_no_access)
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except:
        return HttpResponseBadRequest("Something wrong happened.")

    context = {
        "add_group_form": form
    }

    return HttpResponse(json.dumps({'form': loader.render_to_string(
        "core_main_app/user/workspaces/list/modals/add_group_form.html", context)}),
        'application/javascript')


@login_required
def add_group_right_to_workspace(request):
    """ Add rights to group for the workspace.

    Args:
        request

    Returns
    """
    workspace_id = request.POST.get('workspace_id', None)
    groups_ids = request.POST.getlist('groups_id[]', [])
    is_read_checked = request.POST.get('read', None) == 'true'
    is_write_checked = request.POST.get('write', None) == 'true'

    if len(groups_ids) == 0:
        return HttpResponseBadRequest("You need to select at least one group.")
    if not is_read_checked and not is_write_checked:
        return HttpResponseBadRequest("You need to select at least one permission (read and/or write).")

    try:
        workspace = workspace_api.get_by_id(str(workspace_id))
        for group in group_api.get_all_groups_by_list_id(groups_ids):
            if is_read_checked:
                workspace_api.add_group_read_access_to_workspace(workspace, group, request.user)
            if is_write_checked:
                workspace_api.add_group_write_access_to_workspace(workspace, group, request.user)
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception as exc:
        return HttpResponseBadRequest('Something wrong happened.')

    return HttpResponse(json.dumps({}), content_type='application/javascript')


class AssignView(View):
    """  Assign Ajax view
    """

    api = None

    def post(self, request):
        """ Assign the record to a workspace.

            Args:
                request:

            Returns:
            """
        document_ids = request.POST.getlist('document_id[]', [])
        workspace_id = request.POST.get('workspace_id', None)

        if workspace_id is None or workspace_id == '':
            workspace = None
        else:
            try:
                workspace = workspace_api.get_by_id(str(workspace_id))
            except DoesNotExist:
                return HttpResponseBadRequest("The selected workspace does not exist anymore.")
            except Exception as exc:
                return HttpResponseBadRequest("Something wrong happened.")

        for data_id in document_ids:
            try:
                self.api.assign(self.api.get_by_id(data_id, request.user),
                                workspace,
                                request.user)
            except AccessControlError as ace:
                return HttpResponseBadRequest(escape(str(ace)))
            except Exception as exc:
                return HttpResponseBadRequest("Something wrong happened.")

        return HttpResponse(json.dumps({}), content_type='application/javascript')



""" Menu configuration for core_main_app. Upon installation of the app the following menus are displayed:

  * User menu

    * Home

  * Admin menu

    * Dashboard
    * Link to Django admin
    * Login message
    * Templates (list and upload)
    * XSLT (list and upload)
"""
from django.urls import reverse
from menu import Menu, MenuItem


Menu.add_item(
    "main", MenuItem("Home", reverse("core_main_app_homepage"), icon="home", weight=-1000)
)

Menu.add_item(
    "admin", MenuItem("Dashboard", reverse("admin:core_main_app_admin_home"), icon="tachometer-alt", weight=-90000),
)

Menu.add_item(
    "admin", MenuItem("Django admin", reverse("admin:index"), icon="sitemap", weight=-80000),
)

Menu.add_item(
    "admin", MenuItem("Login message", reverse("admin:core_main_app_login_page"), icon="file-alt", weight=-70000),
)

templates_children = (
    MenuItem("Template List", reverse("admin:core_main_app_templates"), icon="list"),
    MenuItem("Upload New Template", reverse("admin:core_main_app_upload_template"), icon="upload"),
)

Menu.add_item(
    "admin", MenuItem("TEMPLATES", None, children=templates_children)
)

xslt_children = (
    MenuItem("XSLT List", reverse("admin:core_main_app_xslt"), icon="list"),
    MenuItem("Upload New XSLT", reverse("admin:core_main_app_upload_xslt"), icon="upload")
)

Menu.add_item(
    "admin", MenuItem("XSLT", None, children=xslt_children)
)



""" Serializers used throughout the TemplateXslRendering Rest API
"""
from rest_framework_mongoengine.fields import ReferenceField
from rest_framework_mongoengine.serializers import DocumentSerializer

from core_main_app.components.template.models import Template
from core_main_app.components.template_xsl_rendering import api as template_xsl_rendering_api
from core_main_app.components.template_xsl_rendering.models import TemplateXslRendering


class TemplateXslRenderingSerializer(DocumentSerializer):
    """ TemplateXslRendering Serializer
    """
    template = ReferenceField(model=Template)

    class Meta(object):
        """ Meta class
        """
        model = TemplateXslRendering
        fields = "__all__"

    def update(self, instance, validated_data):
        """ Update and return an existing `TemplateXslRendering` instance, given the validated data.
        """
        template_id = validated_data.get("template", None)
        list_xslt = validated_data.get("list_xslt", None)
        detail_xslt = validated_data.get("detail_xslt", None)

        return template_xsl_rendering_api.upsert(
            template_id, list_xslt, detail_xslt, template_xsl_rendering_id=instance.id
        )



""" Fixtures files for Blobs
"""
from core_main_app.components.blob.models import Blob
from core_main_app.components.workspace.models import Workspace
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface


class BlobFixtures(FixtureInterface):
    """ Blob fixtures
    """
    blob_1 = None
    blob_2 = None
    blob_3 = None
    blob_collection = None

    def insert_data(self):
        """ Insert a set of Blobs.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_blob_collection()

    def generate_blob_collection(self):
        """ Generate a Blob collection.

        Returns:

            user 1 -> blob1, blob2
            user 2 -> blob3

        """
        # NOTE: no real file to avoid using unsupported GridFS mock
        self.blob_1 = Blob(filename='blob1',
                           user_id='1',
                           handle='handle1').save()
        self.blob_2 = Blob(filename='blob2',
                           user_id='1',
                           handle='handle2').save()
        self.blob_3 = Blob(filename='blob3',
                           user_id='2',
                           handle='handle3').save()

        self.blob_collection = [self.blob_1,
                                self.blob_2,
                                self.blob_3]


class AccessControlBlobFixture(FixtureInterface):
    """ Access Control Blob fixture
    """
    USER_1_NO_WORKSPACE = 0
    USER_2_NO_WORKSPACE = 1
    USER_1_WORKSPACE_1 = 2
    USER_2_WORKSPACE_2 = 3

    workspace_1 = None
    workspace_2 = None
    workspace_without_data = None
    blob_collection = None

    def insert_data(self):
        """ Insert a set of Blob.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_workspace()
        self.generate_blob_collection()

    def generate_blob_collection(self):
        """ Generate a Blob collection.

        Returns:

        """
        blob_1 = Blob(filename='blob1',
                      user_id='1',
                      handle='handle1').save()
        blob_2 = Blob(filename='blob2',
                      user_id='2',
                      handle='handle2').save()
        blob_3 = Blob(filename='blob3',
                      user_id='1',
                      handle='handle3',
                      workspace=self.workspace_1.id).save()
        blob_4 = Blob(filename='blob4',
                      user_id='2',
                      handle='handle4',
                      workspace=self.workspace_2.id).save()
        self.blob_collection = [blob_1, blob_2, blob_3, blob_4]

    def generate_workspace(self):
        """ Generate workspaces.

        Returns:

        """
        self.workspace_1 = Workspace(title="Workspace 1",
                                     owner='1',
                                     read_perm_id='1',
                                     write_perm_id='1').save()
        self.workspace_2 = Workspace(title="Workspace 2",
                                     owner='2',
                                     read_perm_id='2',
                                     write_perm_id='2').save()
        self.workspace_without_data = Workspace(title="Workspace 3",
                                                owner='3',
                                                read_perm_id='3',
                                                write_perm_id='3').save()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-09-26 14:40
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Main',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'verbose_name': 'core_main_app',
                'permissions': (('publish_data', 'Can publish data'),),
                'default_permissions': (),
            },
        ),
    ]



""" Authentication tests for Web Page REST API
"""
from django.test import SimpleTestCase
from mock.mock import patch
from rest_framework import status

import core_main_app.rest.web_page.views as web_page_views
from core_main_app.components.web_page.models import WebPage
from core_main_app.rest.web_page.serializers import WebPageSerializer
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock


class TestWebPageListGetPermission(SimpleTestCase):

    @patch('core_main_app.components.web_page.api.get')
    def test_anonymous_returns_http_200(self, web_page_api_get):
        web_page_api_get.return_value = _get_mock_web_page()

        response = RequestMock.do_request_get(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            None
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch('core_main_app.components.web_page.api.get')
    def test_authenticated_returns_http_200(self, web_page_api_get):
        web_page_api_get.return_value = _get_mock_web_page()
        mock_user = create_mock_user('1')

        response = RequestMock.do_request_get(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch('core_main_app.components.web_page.api.get')
    def test_staff_returns_http_200(self, web_page_api_get):
        web_page_api_get.return_value = _get_mock_web_page()
        mock_user = create_mock_user('1', is_staff=True)

        response = RequestMock.do_request_get(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestWebPageListPostPermission(SimpleTestCase):

    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_post(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_http_403(self):
        mock_user = create_mock_user('1')

        response = RequestMock.do_request_post(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('core_main_app.components.web_page.api.get')
    @patch.object(WebPageSerializer, "is_valid")
    @patch.object(WebPageSerializer, "save")
    @patch.object(WebPageSerializer, "data")
    def test_staff_returns_http_201(self,
                                    serializer_data,
                                    serializer_save,
                                    serializer_is_valid,
                                    web_page_api_get):
        web_page_api_get.return_value = _get_mock_web_page()
        serializer_data.return_value = True
        serializer_is_valid.return_value = {}
        serializer_save.return_value = None
        mock_user = create_mock_user('1', is_staff=True)

        response = RequestMock.do_request_post(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)


class TestWebPageListDeletePermission(SimpleTestCase):

    def test_anonymous_returns_http_403(self):
        response = RequestMock.do_request_delete(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            None
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_authenticated_returns_http_403(self):
        mock_user = create_mock_user('1')

        response = RequestMock.do_request_delete(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('core_main_app.components.web_page.api.get')
    @patch.object(WebPageSerializer, "is_valid")
    @patch.object(WebPageSerializer, "save")
    @patch.object(WebPageSerializer, "data")
    def test_staff_returns_http_204(self,
                                    serializer_data,
                                    serializer_save,
                                    serializer_is_valid,
                                    web_page_api_get):
        web_page_api_get.return_value = _get_mock_web_page()
        serializer_data.return_value = True
        serializer_is_valid.return_value = {}
        serializer_save.return_value = None
        mock_user = create_mock_user('1', is_staff=True)

        response = RequestMock.do_request_delete(
            web_page_views.WebPageList.as_view(
                web_page_type="login"
            ),
            mock_user
        )

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)


def _get_mock_web_page():
    """ Returns a Web page
    """
    type_default_for_test = 0
    return WebPage(type=type_default_for_test, content="web page")



from __future__ import unicode_literals
from django.db import migrations
from django.contrib.sites.models import Site

from core_main_app.settings import SERVER_URI


def update_site(code, reverse_code):
    site_domain = SERVER_URI

    for i in (("http://", ""), ("https://", "")):
        site_domain = site_domain.replace(*i)

    my_site = Site.objects.get_or_create(id=1)[0]
    my_site.name = site_domain
    my_site.domain = site_domain
    my_site.save()


class Migration(migrations.Migration):

    dependencies = [
        ('sites', '0001_initial'),
        ('core_main_app', '0001_main'),
    ]

    operations = [
        migrations.RunPython(update_site)
    ]



""" REST views for the template version manager API
"""
from django.http import Http404
from django.utils.decorators import method_decorator
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from core_main_app.commons import exceptions as exceptions
from core_main_app.components.template_version_manager import api as template_version_manager_api
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.rest.template_version_manager.abstract_views import AbstractTemplateVersionManagerList, \
    AbstractTemplateList, AbstractStatusTemplateVersion, AbstractStatusTemplateVersionManager, \
    AbstractTemplateVersionManagerDetail
from core_main_app.rest.template_version_manager.serializers import TemplateVersionManagerSerializer, \
    CreateTemplateSerializer
from core_main_app.utils.decorators import api_staff_member_required


class GlobalTemplateVersionManagerList(AbstractTemplateVersionManagerList):
    """ List all GlobalTemplateVersionManager
    """

    def get_template_version_managers(self):
        """ Get GlobalTemplateVersionManager

        Returns:

            List of GlobalTemplateVersionManager
        """
        return template_version_manager_api.get_global_version_managers()


class UserTemplateVersionManagerList(AbstractTemplateVersionManagerList):
    """ List all UserTemplateVersionManager
    """

    permission_classes = (IsAuthenticated, )

    def get_template_version_managers(self):
        """ Get all UserTemplateVersionManager

        Returns:

            List of UserTemplateVersionManager
        """
        return template_version_manager_api.get_all_by_user_id(user_id=str(self.request.user.id))


class TemplateVersionManagerDetail(APIView):
    """ Retrieve a TemplateVersionManager
    """

    permission_classes = (IsAuthenticated,)

    def get_object(self, pk):
        """ Get TemplateVersionManager from db

        Args:

            pk: ObjectId

        Returns:

            TemplateVersionManager
        """
        try:
            return version_manager_api.get(pk)
        except exceptions.DoesNotExist:
            raise Http404

    def get(self, request, pk):
        """ Retrieve a TemplateVersionManager

        Args:

            request: HTTP request
            pk: ObjectId

        Returns:

            - code: 200
              content: TemplateVersionManager
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            template_version_manager_object = self.get_object(pk)

            # Serialize object
            serializer = TemplateVersionManagerSerializer(template_version_manager_object)

            # Return response
            return Response(serializer.data)
        except Http404:
            content = {'message': 'Template Version Manager not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class TemplateVersion(AbstractTemplateVersionManagerDetail):
    """ Create a TemplateVersion
    """

    @method_decorator(api_staff_member_required())
    def post(self, request, pk):
        """ Create a TemplateVersion

        Parameters:

            {
                "filename": "filename",
                "content": "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'><xs:element name='root'/></xs:schema>"
            }

        Note: 
        
            "dependencies_dict": json.dumps({"schemaLocation1": "id1" ,"schemaLocation2":"id2"})

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: Created TemplateVersionManager
            - code: 400
              content: Validation error
            - code: 404
              content: Template was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            template_version_manager_object = self.get_object(pk)

            # Build serializers
            template_serializer = CreateTemplateSerializer(data=request.data)

            # Validate data
            template_serializer.is_valid(True)

            # Save data
            template_serializer.save(template_version_manager=template_version_manager_object)

            return Response(template_serializer.data, status=status.HTTP_201_CREATED)
        except Http404:
            content = {'message': 'Template Version Manager not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class UserTemplateList(AbstractTemplateList):
    """ Create a Template (linked to the user)
    """

    permission_classes = (IsAuthenticated, )

    def post(self, request):
        """ Create a Template (linked to the user)

        Parameters:

            {
                "title": "title",
                "filename": "filename",
                "content": "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'><xs:element name='root'/></xs:schema>"
            }

        Note:

            "dependencies_dict": json.dumps({"schemaLocation1": "id1" ,"schemaLocation2":"id2"})

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Created Template
            - code: 400
              content: Validation error / not unique / XSD error
            - code: 500
              content: Internal server error
        """
        return super(UserTemplateList, self).post(request)

    def get_user(self):
        """ Retrieve the user from the request

        Returns:

            User ID
        """
        return str(self.request.user.id)


class GlobalTemplateList(AbstractTemplateList):
    """ Create a Template (global schema)
    """

    @method_decorator(api_staff_member_required())
    def post(self, request):
        """ Create a Template (global schema)

        Parameters:

            {
                "title": "title",
                "filename": "filename",
                "content": "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'><xs:element name='root'/></xs:schema>"
            }

        Note:

            "dependencies_dict": json.dumps({"schemaLocation1": "id1" ,"schemaLocation2":"id2"})

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Created template
            - code: 400
              content: Validation error / not unique / XSD error
            - code: 500
              content: Internal server error
        """
        return super(GlobalTemplateList, self).post(request)

    def get_user(self):
        """ The user is None for a global template

        Returns:

            None
        """
        return None


class CurrentTemplateVersion(AbstractStatusTemplateVersion):
    """ Update status to current
    """

    permission_classes = (IsAuthenticated, )

    def status_update(self, template_object):
        """ Update status to current

        Args:

            template_object: template_version

        Returns:

            TemplateVersion
        """
        return version_manager_api.set_current(template_object)


class DisableTemplateVersion(AbstractStatusTemplateVersion):
    """ Update status to disabled
    """

    permission_classes = (IsAuthenticated,)

    def status_update(self, template_object):
        """ Update status to disabled

        Args:

            template_object: template_version

        Returns:

            TemplateVersion
        """
        return version_manager_api.disable_version(template_object)


class RestoreTemplateVersion(AbstractStatusTemplateVersion):
    """ Update status to restored
    """

    permission_classes = (IsAuthenticated,)

    def status_update(self, template_object):
        """ Update status to restored

        Args:

            template_object: template_version

        Returns:

            TemplateVersion
        """
        return version_manager_api.restore_version(template_object)


class DisableTemplateVersionManager(AbstractStatusTemplateVersionManager):
    """ Update status to disabled
    """

    permission_classes = (IsAuthenticated,)

    def status_update(self, template_version_manager_object):
        """ Update status to disabled

        Args:

            template_version_manager_object: template_version_manager

        Returns:

            TemplateVersionManager
        """
        # FIXME: add return?
        version_manager_api.disable(template_version_manager_object)


class RestoreTemplateVersionManager(AbstractStatusTemplateVersionManager):
    """ Update status to restored
    """

    permission_classes = (IsAuthenticated,)

    def status_update(self, template_version_manager_object):
        """ Update status to restored

        Args:

            template_version_manager_object: template_version_manager

        """
        # FIXME: add return?
        version_manager_api.restore(template_version_manager_object)



""" Urls utils
"""

import re

from django.urls import reverse


def get_template_download_pattern():
    """ Return regex pattern to match an url to download a template.

    Returns:

    """
    # build django url to download a template
    download_template_url = reverse('core_main_app_rest_template_download', kwargs={'pk': 'template_id'})
    # make url a regex
    download_template_regex = download_template_url.replace('template_id', '(?P<pk>\w+)')
    # compile regex
    pattern = re.compile(download_template_regex)

    return pattern


def get_blob_download_regex():
    """ Return regex pattern to match an url to download a blob.

    Returns:

    """
    # build django url to download a blob
    download_blob_url = reverse("core_main_app_rest_blob_download", kwargs={'pk': 'blob_id'})
    download_blob_url = download_blob_url.replace('blob_id/', '')
    # make the regex
    regex = ">(http[s]?:[^<>]+" + download_blob_url + "[0-9a-f]{24}/?)<"
    return regex



""" Template Version Manager unit tests
"""
from unittest.case import TestCase

from bson.objectid import ObjectId
from django.core import exceptions as django_exceptions
from django.test import override_settings
from mock.mock import Mock, patch, MagicMock
from mongoengine import errors as mongoengine_errors

from core_main_app.commons.exceptions import NotUniqueError, ModelError
from core_main_app.components.template.models import Template
from core_main_app.components.template_version_manager import api as version_manager_api
from core_main_app.components.template_version_manager.models import TemplateVersionManager


class TestTemplateVersionManagerInsert(TestCase):
    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch('core_main_app.components.template_version_manager.models.TemplateVersionManager.save')
    @patch('core_main_app.components.template.models.Template.save')
    def test_create_version_manager_returns_version_manager(self, mock_save_template,
                                                            mock_save_template_version_manager):
        # Arrange
        template_filename = "schema.xsd"
        template_content = "<schema xmlns='http://www.w3.org/2001/XMLSchema'></schema>"
        template = _create_template(template_filename, template_content)

        mock_save_template.return_value = template

        version_manager = _create_template_version_manager(title="Schema")
        mock_save_template_version_manager.return_value = version_manager

        # Act
        result = version_manager_api.insert(version_manager, template)

        # Assert
        self.assertIsInstance(result, TemplateVersionManager)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch('core_main_app.components.template.models.Template.delete')
    @patch('core_main_app.components.template.models.Template.save')
    @patch('core_main_app.components.template_version_manager.models.TemplateVersionManager.save')
    def test_insert_manager_raises_api_error_if_title_already_exists(self, mock_version_manager_save,
                                                                     mock_template_save,
                                                                     mock_template_delete):
        # Arrange
        template_filename = "schema.xsd"
        template_content = "<schema xmlns='http://www.w3.org/2001/XMLSchema'></schema>"
        template = _create_template(template_filename, template_content)

        mock_template_save.return_value = template
        mock_template_delete.return_value = None
        mock_version_manager = _create_template_version_manager(title="Schema")
        mock_version_manager_save.side_effect = mongoengine_errors.NotUniqueError("")

        # Act + Assert
        with self.assertRaises(NotUniqueError):
            version_manager_api.insert(mock_version_manager, template)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch('core_main_app.components.template.models.Template.save')
    def test_create_version_manager_raises_exception_if_error_in_create_template(self, mock_save):
        # Arrange
        template_filename = "schema.xsd"
        template_content = "<schema xmlns='http://www.w3.org/2001/XMLSchema'></schema>"
        template = _create_template(template_filename, template_content)

        mock_version_manager = _create_mock_template_version_manager(title="Schema")
        mock_save.side_effect = django_exceptions.ValidationError("")

        # Act + Assert
        with self.assertRaises(django_exceptions.ValidationError):
            version_manager_api.insert(mock_version_manager, template)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch('core_main_app.components.template.models.Template.delete')
    @patch('core_main_app.components.template_version_manager.models.TemplateVersionManager.save')
    @patch('core_main_app.components.template.models.Template.save')
    def test_create_version_manager_raises_exception_if_error_in_create_version_manager(self, mock_save_template,
                                                                                        mock_save_version_manager,
                                                                                        mock_delete_template):
        # Arrange
        template_filename = "Schema"
        template_content = "<schema xmlns='http://www.w3.org/2001/XMLSchema'></schema>"
        template = _create_template(template_filename, template_content)

        mock_save_template.return_value = template
        version_manager = _create_template_version_manager(title="Schema")
        mock_save_version_manager.side_effect = django_exceptions.ValidationError("")
        mock_delete_template.return_value = None

        # Act + Assert
        with self.assertRaises(ModelError):
            version_manager_api.insert(version_manager, template)


class TestTemplateVersionManagerAddVersion(TestCase):
    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch('core_main_app.components.template_version_manager.models.TemplateVersionManager.insert')
    @patch('core_main_app.components.template.models.Template.save')
    @patch('core_main_app.components.template_version_manager.models.TemplateVersionManager.save')
    def test_insert_returns_template_version(self, mock_save_template_version_manager, mock_save_template, mock_insert):
        # Arrange
        template_filename = "Schema"
        template_content = "<schema xmlns='http://www.w3.org/2001/XMLSchema'></schema>"

        template = _create_template(template_filename, template_content)
        mock_save_template.return_value = template
        version_manager = _create_template_version_manager()
        mock_save_template_version_manager.return_value = version_manager
        mock_insert.return_value = version_manager
        # Act
        result = version_manager_api.insert(version_manager, template)

        # Assert
        self.assertIsInstance(result, TemplateVersionManager)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch('core_main_app.components.template.models.Template.save')
    def test_insert_raises_exception_if_error_in_create_template(self, mock_save_template):
        # Arrange
        template_filename = "schema.xsd"
        template_content = "<schema xmlns='http://www.w3.org/2001/XMLSchema'></schema>"
        template = _create_template(template_filename, template_content)

        mock_save_template.side_effect = django_exceptions.ValidationError("")

        mock_version_manager = _create_mock_template_version_manager()

        # Act + Assert
        with self.assertRaises(django_exceptions.ValidationError):
            version_manager_api.insert(mock_version_manager, template)


class TestTemplateVersionManagerGetGlobalVersions(TestCase):

    @patch('core_main_app.components.template_version_manager.models.TemplateVersionManager.'
           'get_global_version_managers')
    def test_get_global_version_managers_returns_templates(self, mock_get_global_version_managers):
        # Arrange
        mock_template1 = _create_mock_template()
        mock_template2 = _create_mock_template()

        mock_get_global_version_managers.return_value = [mock_template1, mock_template2]

        result = version_manager_api.get_global_version_managers()

        # Assert
        self.assertTrue(all(isinstance(item, Template) for item in result))


class TestTemplateVersionManagerGetActiveGlobalVersions(TestCase):

    @patch('core_main_app.components.version_manager.models.VersionManager.get_active_global_version_manager')
    def test_get_active_global_version_managers_returns_templates_not_disable(self,
                                                                              mock_get_active_global_version_managers):
        # Arrange
        mock_template1 = _create_template_version_manager("test1")
        mock_template2 = _create_template_version_manager("test2")

        queryset = MagicMock()
        queryset.filter.return_value = queryset
        queryset.all.return_value = [mock_template1, mock_template2]
        mock_get_active_global_version_managers.return_value = queryset

        result = version_manager_api.get_active_global_version_manager()

        # Assert
        self.assertTrue(all(item.is_disabled is False for item in result))


class TestTemplateVersionManagerGetActiveGlobalVersionsByUserId(TestCase):

    @patch('core_main_app.components.version_manager.models.VersionManager.get_active_version_manager_by_user_id')
    def test_get_active_global_version_managers_by_user_id_returns_templates_not_disable_with_given_user_id(
                                                                            self,
                                                                            mock_get_active_global_version_managers):
        # Arrange
        user_id = "10"
        mock_template1 = _create_template_version_manager(user_id=user_id)
        mock_template2 = _create_template_version_manager(user_id=user_id)

        queryset = MagicMock()
        queryset.filter.return_value = queryset
        queryset.all.return_value = [mock_template1, mock_template2]
        mock_get_active_global_version_managers.return_value = queryset

        result = version_manager_api.get_active_version_manager_by_user_id(user_id)

        # Assert
        self.assertTrue(all(item.is_disabled is False and item.user == str(user_id) for item in result))


class TestTemplateVersionManagerGetDisableGlobalVersions(TestCase):

    @patch('core_main_app.components.version_manager.models.VersionManager.get_disable_global_version_manager')
    def test_get_disable_global_version_managers_returns_templates_disabled(self,
                                                                            mock_get_disable_global_version_managers):
        # Arrange
        mock_template1 = _create_template_version_manager(is_disabled=True)
        mock_template2 = _create_template_version_manager(is_disabled=True)

        queryset = MagicMock()
        queryset.filter.return_value = queryset
        queryset.all.return_value = [mock_template1, mock_template2]
        mock_get_disable_global_version_managers.return_value = queryset

        result = version_manager_api.get_disable_global_version_manager()

        # Assert
        self.assertTrue(all(item.is_disabled is True for item in result))


class TestTemplateVersionManagerGetDisableGlobalVersionsByUserId(TestCase):

    @patch('core_main_app.components.version_manager.models.VersionManager.get_disable_version_manager_by_user_id')
    def test_get_disable_global_version_managers_by_user_id_returns_templates_disabled_with_giver_user_id(
                                                                            self,
                                                                            mock_get_disable_global_version_managers):
        # Arrange
        user_id = "10"
        mock_template1 = _create_template_version_manager(is_disabled=True, user_id=user_id)
        mock_template2 = _create_template_version_manager(is_disabled=True, user_id=user_id)

        queryset = MagicMock()
        queryset.filter.return_value = queryset
        queryset.all.return_value = [mock_template1, mock_template2]
        mock_get_disable_global_version_managers.return_value = queryset

        result = version_manager_api.get_disable_version_manager_by_user_id(user_id)

        # Assert
        self.assertTrue(all(item.is_disabled is True and item.user == str(user_id) for item in result))


class TestTemplateVersionManagerGetAllByVersionIds(TestCase):

    @patch.object(TemplateVersionManager, 'get_all_by_version_ids')
    def test_get_all_by_version_ids(self, mock_get_all_by_version_ids):
        # Arrange
        version_x = ObjectId()
        version_y = ObjectId()
        mock_template1 = _create_mock_template_version_manager("mock1", [version_x])
        mock_template2 = _create_mock_template_version_manager("mock2", [version_y])

        mock_get_all_by_version_ids.return_value = [mock_template1, mock_template2]

        result = version_manager_api.get_all_by_version_ids([version_x, version_y])

        # Assert
        self.assertTrue(all(isinstance(item, TemplateVersionManager) for item in result))


def _create_mock_template(mock_template_filename="", mock_template_content=""):
    """
    Returns a mock template
    :param mock_template_filename:
    :param mock_template_content:
    :return:
    """
    mock_template = Mock(spec=Template)
    mock_template.filename = mock_template_filename
    mock_template.content = mock_template_content
    mock_template.id = ObjectId()
    return mock_template


def _create_mock_template_version_manager(title="", versions=None, is_disabled=False, user_id=""):
    """
    Returns a mock template version manager
    :return:
    """
    if versions is None:
        versions = []
    mock_template_version_manager = Mock(spec=TemplateVersionManager)
    mock_template_version_manager.title = title
    mock_template_version_manager.id = ObjectId()
    mock_template_version_manager.versions = versions
    mock_template_version_manager.disabled_versions = []
    mock_template_version_manager.is_disabled = is_disabled
    mock_template_version_manager.user = str(user_id)
    mock_template_version_manager._cls = TemplateVersionManager.class_name
    return mock_template_version_manager


def _create_template(filename="", content=""):
    """
    Returns a template
    :param filename:
    :param content:
    :return:
    """
    return Template(
        id=ObjectId(),
        filename=filename,
        content=content
    )


def _create_template_version_manager(title="", is_disabled=False, user_id=""):
    """
    Returns a templates version manager
    :param title:
    :return:
    """
    return TemplateVersionManager(
        id=ObjectId(),
        title=title,
        versions=[],
        user=user_id,
        is_disabled=is_disabled,
        disabled_versions=[],
        _cls=TemplateVersionManager.class_name
    )



""" Authentication tests for Template REST API
"""
from django.test import SimpleTestCase
from mock.mock import patch
from rest_framework import status

from core_main_app.components.template.models import Template
from core_main_app.rest.template import views as template_views
from core_main_app.rest.template.serializers import TemplateSerializer
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock


class TestTemplateDetailGetPermission(SimpleTestCase):

    def setUp(self):
        self.fake_id = '507f1f77bcf86cd799439011'

    @patch.object(Template, "get_by_id")
    @patch.object(TemplateSerializer, "data")
    def test_anonymous_returns_http_200(self, template_serializer_data, template_get_by_id):
        template_get_by_id.return_value = {}
        template_serializer_data.return_value = True

        response = RequestMock.do_request_get(
            template_views.TemplateDetail.as_view(),
            None,
            param={'pk': self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Template, "get_by_id")
    @patch.object(TemplateSerializer, "data")
    def test_authenticated_returns_http_200(self, template_serializer_data, template_get_by_id):
        template_get_by_id.return_value = {}
        template_serializer_data.return_value = True

        mock_user = create_mock_user('1')

        response = RequestMock.do_request_get(
            template_views.TemplateDetail.as_view(),
            mock_user,
            param={'pk': self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Template, "get_by_id")
    @patch.object(TemplateSerializer, "data")
    def test_staff_returns_http_200(self, template_serializer_data, template_get_by_id):
        template_get_by_id.return_value = {}
        template_serializer_data.return_value = True

        mock_user = create_mock_user('1', is_staff=True)

        response = RequestMock.do_request_get(
            template_views.TemplateDetail.as_view(),
            mock_user,
            param={'pk': self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestTemplateDownloadGetPermission(SimpleTestCase):

    def setUp(self):
        self.fake_id = '507f1f77bcf86cd799439011'

    @patch.object(Template, "get_by_id")
    def test_anonymous_returns_http_200(self, template_get_by_id):
        template_get_by_id.return_value = Template(content='test', filename='test.txt')

        response = RequestMock.do_request_get(
            template_views.TemplateDownload.as_view(),
            None,
            param={'pk': self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Template, "get_by_id")
    def test_authenticated_returns_http_200(self, template_get_by_id):
        template_get_by_id.return_value = Template(content='test', filename='test.txt')

        mock_user = create_mock_user('1')

        response = RequestMock.do_request_get(
            template_views.TemplateDownload.as_view(),
            mock_user,
            param={'pk': self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch.object(Template, "get_by_id")
    def test_staff_returns_http_200(self, template_get_by_id):
        template_get_by_id.return_value = Template(content='test', filename='test.txt')

        mock_user = create_mock_user('1', is_staff=True)

        response = RequestMock.do_request_get(
            template_views.TemplateDownload.as_view(),
            mock_user,
            param={'pk': self.fake_id}
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)



""" Core main app user views
"""
from django.conf import settings
from django.contrib.auth import authenticate, login, logout, get_user_model
from django.contrib.auth.decorators import login_required
from django.contrib.auth.forms import (
    PasswordResetForm, SetPasswordForm)
from django.contrib.auth.tokens import default_token_generator
from django.contrib.staticfiles import finders
from django.http.response import HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect, resolve_url
from django.urls import reverse
from django.urls import reverse_lazy
from django.utils.encoding import force_text
from django.utils.http import urlsafe_base64_decode
from rest_framework.status import HTTP_405_METHOD_NOT_ALLOWED

import core_main_app.components.web_page_login.api as web_page_login_api
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.settings import INSTALLED_APPS
from core_main_app.utils.markdown_parser import parse
from core_main_app.utils.rendering import render
from core_main_app.views.user.forms import LoginForm

if "defender" in INSTALLED_APPS:
    from defender.decorators import watch_login

    @watch_login()
    def defender_custom_login(request):
        """ Custom login page with defender controls.

        Args:
            request:

        Returns:

        """
        return default_custom_login(request)


def custom_login(request):
    """ Custom login page calls default page or page with defender if installed.

    Args:
        request:

    Returns:

    """
    if "defender" in INSTALLED_APPS:
        return defender_custom_login(request)
    else:
        return default_custom_login(request)


def default_custom_login(request):
    """ Default custom login page.
    
        Parameters:
            request: 
    
        Returns:
    """

    def _login_redirect(to_page):
        if to_page is not None and to_page != "":
            return redirect(to_page)

        return redirect(reverse("core_main_app_homepage"))

    if request.method == "POST":
        username = request.POST["username"]
        password = request.POST["password"]
        next_page = request.POST["next_page"]

        try:
            user = authenticate(username=username, password=password)

            if not user.is_active:
                return render(
                    request, "core_main_app/user/login/main.html", context={
                        "login_form": LoginForm(
                            initial={"next_page": next_page}
                        ),
                        "login_locked": True,
                        "with_website_features":
                            "core_website_app" in INSTALLED_APPS
                    }
                )

            login(request, user)

            return _login_redirect(next_page)
        except Exception:
            return render(
                request, "core_main_app/user/login/main.html", context={
                    "login_form": LoginForm(initial={"next_page": next_page}),
                    "login_error": True,
                    "with_website_features":
                        "core_website_app" in INSTALLED_APPS
                }
            )
    elif request.method == "GET":
        if request.user.is_authenticated:
            return redirect(reverse("core_main_app_homepage"))

        next_page = None
        if "next" in request.GET:
            next_page = request.GET["next"]

        # build the context
        context = {
            "login_form": LoginForm(initial={"next_page": next_page}),
            "with_website_features": "core_website_app" in INSTALLED_APPS
        }
        assets = {
            "css": ["core_main_app/user/css/login.css"]
        }

        # get the web page login if exist
        web_page_login = web_page_login_api.get()

        if web_page_login:
            # update the context
            context["login_message"] = parse(web_page_login.content)

            # if exist we build assets and modals collection
            assets.update({
                "js": [{
                    "path": "core_main_app/user/js/login/message.js",
                    "is_raw": False
                }]
            })
            modals = ["core_main_app/user/login/modals/login_message.html"]

            # render the page with context, assets and modals
            return render(request,
                          "core_main_app/user/login/main.html",
                          context=context,
                          assets=assets,
                          modals=modals)

        # render the page with context, assets
        return render(request,
                      "core_main_app/user/login/main.html",
                      context=context,
                      assets=assets)
    else:
        return HttpResponse(status=HTTP_405_METHOD_NOT_ALLOWED)


def custom_logout(request):
    """ Custom logout page.
    
        Parameters:
            request: 
    
        Returns:
    """
    logout(request)
    return redirect(reverse("core_main_app_homepage"))


def homepage(request):
    """ Homepage for the website

        Parameters:
            request:

        Returns:
    """
    assets = dict()

    if finders.find("core_main_app/css/homepage.css") is not None:
        assets["css"] = ["core_main_app/css/homepage.css"]

    if finders.find("core_main_app/js/homepage.js") is not None:
        assets["js"] = [{
            "path": "core_main_app/js/homepage.js",
            "is_raw": False
        }]

    return render(request, "core_main_app/user/homepage.html", assets=assets)


@login_required(login_url=reverse_lazy("core_main_app_login"))
def manage_template_versions(request, version_manager_id):
    """View that allows template versions management

    Args:
        request:
        version_manager_id:

    Returns:

    """
    try:
        # get the version manager
        version_manager = version_manager_api.get(version_manager_id)

        if not request.user.is_staff and version_manager.user != str(request.user.id):
            raise Exception("You do not have the rights to perform this action.")

        context = get_context_manage_template_versions(version_manager)
        if "core_parser_app" in settings.INSTALLED_APPS:
            context.update({"module_url": "core_parser_app_template_modules"})

        assets = {
            "js": [
                {
                    "path": "core_main_app/common/js/templates/versions/set_current.js",
                    "is_raw": False
                },
                {
                    "path": "core_main_app/common/js/templates/versions/restore.js",
                    "is_raw": False
                },
                {
                    "path": "core_main_app/common/js/templates/versions/modals/disable.js",
                    "is_raw": False
                }
            ]
        }

        modals = ["core_main_app/admin/templates/versions/modals/disable.html"]

        return render(request,
                      "core_main_app/common/templates/versions.html",
                      assets=assets,
                      modals=modals,
                      context=context)
    except Exception as e:
        return render(request,
                      "core_main_app/common/commons/error.html",
                      context={"error": str(e)})


def get_context_manage_template_versions(version_manager, object_name="Template"):
    """ Get the context to manage the template versions.

    Args:
        version_manager:
        object_name:

    Returns:

    """

    # Use categorized version for easier manipulation in template
    versions = version_manager.versions
    categorized_versions = {
        "available": [],
        "disabled": []
    }
    for index, version in enumerate(versions, 1):
        indexed_version = {
            "index": index,
            "object": version
        }

        if version not in version_manager.disabled_versions:
            categorized_versions["available"].append(indexed_version)
        else:
            categorized_versions["disabled"].append(indexed_version)

    version_manager.versions = categorized_versions
    context = {
        "object_name": object_name,
        "version_manager": version_manager
    }

    return context


def custom_reset_password(request,
                          template_name="core_main_app/user/registration/password_reset_form.html",
                          email_template_name="core_main_app/user/registration/password_reset_email.html",
                          subject_template_name="core_main_app/user/registration/password_reset_subject.txt",
                          password_reset_form=PasswordResetForm,
                          token_generator=default_token_generator,
                          post_reset_redirect=None,
                          from_email=None,
                          extra_context=None,
                          html_email_template_name=None,
                          extra_email_context=None):
    """ Custom reset password page.

        Parameters:
        :param request:
        :param template_name:
        :param email_template_name:
        :param subject_template_name:
        :param password_reset_form:
        :param token_generator:
        :param post_reset_redirect:
        :param from_email:
        :param extra_context:
        :param extra_email_context:
        :param html_email_template_name:

        Returns:
        :return request
    """

    if post_reset_redirect is None:
        post_reset_redirect = reverse("password_reset_done")
    else:
        post_reset_redirect = resolve_url(post_reset_redirect)
    if request.method == "POST":
        form = password_reset_form(request.POST)
        if form.is_valid():
            opts = {
                "use_https": request.is_secure(),
                "token_generator": token_generator,
                "from_email": from_email,
                "email_template_name": email_template_name,
                "subject_template_name": subject_template_name,
                "request": request,
                "html_email_template_name": html_email_template_name,
                "extra_email_context": extra_email_context,
            }
            form.save(**opts)
            return HttpResponseRedirect(post_reset_redirect)
    else:
        form = password_reset_form()
    context = {
        "form": form,
        "title": "Password reset",
    }
    if extra_context is not None:
        context.update(extra_context)

    return render(request, template_name, context=context)


def custom_password_reset_done(request,
                               template_name="core_main_app/user/registration/password_reset_done.html",
                               extra_context=None):
    """ Custom password reset done page.

        Parameters:
        :param request:
        :param template_name:
        :param extra_context:

        Returns:
        :return request
    """

    context = {
        "title": "Password reset sent",
    }

    assets = {
        "css": ["core_main_app/user/css/registration.css"]
    }

    if extra_context is not None:
        context.update(extra_context)

    return render(request, template_name, assets=assets, context=context)


def custom_password_reset_confirm(request, uidb64=None, token=None,
                                  template_name="core_main_app/user/registration/password_reset_confirm.html",
                                  token_generator=default_token_generator,
                                  set_password_form=SetPasswordForm,
                                  post_reset_redirect=None, extra_context=None):
    """
    View that checks the hash in a password reset link and presents a
    form for entering a new password.

        Parameters:
        :param request:
        :param uidb64:
        :param token:
        :param template_name:
        :param token_generator:
        :param set_password_form:
        :param post_reset_redirect:
        :param extra_context:

        Returns:
        :return request
    """

    user_model = get_user_model()

    assert uidb64 is not None and token is not None  # checked by URLconf
    if post_reset_redirect is None:
        post_reset_redirect = reverse("password_reset_complete")
    else:
        post_reset_redirect = resolve_url(post_reset_redirect)
    try:
        # urlsafe_base64_decode() decodes to bytestring on Python 3
        uid = force_text(urlsafe_base64_decode(uidb64))
        user = user_model._default_manager.get(pk=uid)
    except (TypeError, ValueError, OverflowError, user_model.DoesNotExist):
        user = None

    if user is not None and token_generator.check_token(user, token):
        validlink = True
        title = "Enter new password"
        if request.method == "POST":
            form = set_password_form(user, request.POST)
            if form.is_valid():
                form.save()
                return HttpResponseRedirect(post_reset_redirect)
        else:
            form = set_password_form(user)
    else:
        validlink = False
        form = None
        title = "Password reset unsuccessful"
    context = {
        "form": form,
        "title": title,
        "validlink": validlink,
    }
    if extra_context is not None:
        context.update(extra_context)

    return render(request, template_name, context=context)


def custom_password_reset_complete(request,
                                   template_name="core_main_app/user/registration/password_reset_complete.html",
                                   extra_context=None):
    """
    Custom password reset complete page.

        Parameters:
        :param request:
        :param template_name:
        :param extra_context:

        Returns:
        :return request
    """
    context = {
        "login_url": resolve_url(settings.LOGIN_URL),
        "title": "Password reset complete",
    }

    assets = {
        "css": ["core_main_app/user/css/registration.css"]
    }

    if extra_context is not None:
        context.update(extra_context)

    return render(request, template_name, context=context, assets=assets)



""" Web Page serializer
"""
from rest_framework_mongoengine.serializers import DocumentSerializer

import core_main_app.components.web_page.api as web_page_api
from core_main_app.components.web_page.models import WebPage


class WebPageSerializer(DocumentSerializer):
    """ Represents the web page serializer
    """
    class Meta(object):
        model = WebPage

        fields = ["id",
                  "type",
                  "content"]

        read_only_fields = ('id', )

    def create(self, validated_data):
        """ Create and return a new `WebPage` instance, given the validated data
        """
        # Create data
        web_page = WebPage(
            content=validated_data['content'],
            type=validated_data['type'],
        )
        # Save the web page
        return web_page_api.upsert(web_page)

    def update(self, instance, validated_data):
        """ Update and return an existing `WebPage` instance, given the validated data
        """
        instance.content = validated_data.get('content', instance.content)
        return web_page_api.upsert(instance)



""" Parser to convert Markdown to HTML in a safe way
"""
from markdown import markdown

from xml_utils.commons.exceptions import HTMLError
from xml_utils.html_tree.parser import safe_html


def parse(text):
    """ Parse Markdown to convert it into HTML

    Args:
        text:

    Returns:
    """
    md_text = markdown(text)

    try:
        return safe_html(md_text)
    except HTMLError as e:
        return str(e)



""" Templatetags for core_website_app
"""
import re

from django import template
from django.template.defaultfilters import stringfilter
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter(name="stripjs")
@stringfilter
def stripjs(value):
    """ Removes the javascript tags susceptible to be executed on the client side

    Args:
        value:

    Returns:

    """
    stripped = re.sub(r'<script(?:\s[^>]*)?(>(?:.(?!/script>))*</script>|/>)', '', force_text(value), flags=re.S)
    return mark_safe(stripped)



""" REST abstract views for the data API
"""
import json
from abc import ABCMeta, abstractmethod

from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from core_main_app.components.data import api as data_api
from core_main_app.utils.query.constants import VISIBILITY_OPTION
from core_main_app.utils.query.mongo.query_builder import QueryBuilder


class AbstractExecuteLocalQueryView(APIView, metaclass=ABCMeta):
    sub_document_root = 'dict_content'

    def get(self, request):
        """ Execute query on local instance and return results

        Parameters:

            # get all results (paginated)
            {"query": "{}"}
            # get all results
            {"query": "{}", "all": "true"}
            # get all results filtered by title
            {"query": "{}", "title": "title_string"}
            # get all results filtered by workspaces
            {"query": "{}", "workspaces": "[{\\"id\\":\\"[workspace_id]\\"}]"}
            # get all results filtered by private workspace
            {"query": "{}", "workspaces": "[{\\"id\\":\\"None\\"}]"}
            # get all results filtered by templates
            {"query": "{}", "templates": "[{\\"id\\":\\"[template_id]\\"}]"}
            # get all results that verify a given criteria
            {"query": "{\\"root.element.value\\": 2}"}
            # get results using multiple options
            {"query": "{\\"root.element.value\\": 2}", "workspaces": "[{\\"id\\":\\"workspace_id\\"}]", "all": "true"}
            {"query": "{\\"root.element.value\\": 2}", "templates": "[{\\"id\\":\\"template_id\\"}]", "all": "true"}
            {"query": "{\\"root.element.value\\": 2}", "templates": "[{\\"id\\":\\"template_id\\"}]",
            "workspaces": "[{\\"id\\":\\"[workspace_id]\\"}]","all": "true"}


        Warning:

            Need to backslash double quotes in JSON payload

        Args:

            request: HTTP request

        Returns:

            - code: 200
              content: List of data
            - code: 400
              content: Bad request
            - code: 500
              content: Internal server error
        """
        return self.execute_query()

    def post(self, request):
        """ Execute query on local instance and return results

        Parameters:

            {"query": "{\"$or\": [{\"image.owner\": \"Peter\"}, {\"image.owner.#text\":\"Peter\"}]}"}

        Warning:

            Need to backslash double quotes in JSON payload

        Args:
            request:

        Returns:

            - code: 200
              content: List of data
            - code: 400
              content: Bad request
            - code: 500
              content: Internal server error
        """
        return self.execute_query()

    def execute_query(self):
        """ Compute and return query results
        """
        try:
            # get query and templates
            query = self.request.data.get('query', None)
            templates = json.loads(self.request.data.get('templates', '[]'))
            workspaces = json.loads(self.request.data.get('workspaces', '[]'))
            options = json.loads(self.request.data.get('options', '{}'))
            title = self.request.data.get('title', None)
            order_by_field = self.request.data.get('order_by_field', '').split(',')
            if query is not None:
                # prepare query
                raw_query = self.build_query(query=query,
                                             templates=templates,
                                             options=options,
                                             workspaces=workspaces,
                                             title=title)
                # execute query
                data_list = self.execute_raw_query(raw_query, order_by_field)
                # build and return response
                return self.build_response(data_list)
            else:
                content = {'message': 'Expected parameters not provided.'}
                return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def build_query(self, query, workspaces=None, templates=None, options=None, title=None):
        """ Build the raw query

        Args:

            query: Query
            workspaces: List of workspace
            templates: List of template
            options: Query option
            title: title filter

        Returns:

            The raw query
        """

        # build query builder
        query_builder = QueryBuilder(query, self.sub_document_root)
        # update the criteria with workspaces information
        if workspaces is not None and len(workspaces) > 0:
            list_workspace_ids = [workspace['id'] for workspace in workspaces]
            query_builder.add_list_criteria('workspace', list_workspace_ids)
        # update the criteria with templates information
        if templates is not None and len(templates) > 0:
            list_template_ids = [template['id'] for template in templates]
            query_builder.add_list_criteria('template', list_template_ids)
        # update the criteria with visibility information
        if options is not None and VISIBILITY_OPTION in options:
            query_builder.add_visibility_criteria(options[VISIBILITY_OPTION])
        # update the criteria with title information
        if title is not None:
            query_builder.add_title_criteria(title)

        # get raw query
        return query_builder.get_raw_query()

    def execute_raw_query(self, raw_query, order_by_field):
        """ Execute the raw query in database

        Args:

            raw_query: Query to execute
            order_by_field:

        Returns:

            Results of the query
        """
        return data_api.execute_query(raw_query, self.request.user, order_by_field)

    @abstractmethod
    def build_response(self, data_list):
        """ Build the paginated response.

        Args:

            data_list: List of data.

        Returns:

            The response
        """
        raise NotImplementedError("build_response method is not implemented.")


""" Logger utils
"""
import copy


def update_logger_with_local_app(logging_dict, local_logger_conf, list_app):
    """ Update the logger with local app.

    Args:
        logging_dict:
        local_logger_conf:
        list_app:
    Returns:
    """
    logging_dict['loggers'].update({app: copy.deepcopy(local_logger_conf) for app in get_list_local_app(list_app)})


def get_list_local_app(list_app):
    """ Get the list of local apps.

    Args:
         list_app:
    Return:
    """
    local_apps = []
    for app in list_app:
        if app.startswith("core") and "." not in app:
            local_apps.append(app)
    return local_apps


def set_generic_handler(logging_dict, name, level, filename, max_size, backup_count, class_name):
    """ Set a handler based on the parameters.

    Args
        logging_dict:
        name:
        level:
        filename:
        max_size:
        backup_count:
        class_name:

    Returns:
    """
    logging_dict['handlers'].update({
        name: {
            'level': level,
            'class': class_name,
            'filename': filename,
            'maxBytes': max_size,
            'backupCount': backup_count,
            'formatter': 'fmt-default',
        },
    })


def set_generic_logger(logging_dict, name, level, list_handler):
    """ Set a logger based on the parameters.

    Args
        logging_dict:
        name:
        level:
        list_handler:

    Returns:
    """
    logging_dict['loggers'].update({
        name: {
            'handlers': list_handler,
            'level': level,
            'propagate': False,
        },
    })



""" LXML URI Resolver utils
"""
from core_main_app.commons.exceptions import CoreError
from core_main_app.settings import XSD_URI_RESOLVER
from core_main_app.utils.resolvers.xsd_uri_resolvers import XSD_URI_RESOLVERS


def lmxl_uri_resolver():
    """ Return lxml uri resolver according to the settings

    Returns:

    """
    uri_resolver = None

    if XSD_URI_RESOLVER:
        # Check that the setting is an accepted value
        if XSD_URI_RESOLVER not in XSD_URI_RESOLVERS:
            raise CoreError("Error: XSD_URI_RESOLVER setting has an incorrect value.")

        # Return the correct resolver depending on the setting
        uri_resolver = XSD_URI_RESOLVERS[XSD_URI_RESOLVER]()

    return uri_resolver



""" lxml URI resolver using requests api
"""
import logging

from core_main_app.utils.requests_utils.requests_utils import send_get_request
from xml_utils.xml_validation.resolvers.default_uri_resolver import DefaultURIResolver

logger = logging.getLogger(__name__)


class RequestsResolver(DefaultURIResolver):
    """ Requests URI Resolver for lxml
    """
    def resolve(self, url, id, context):
        """ Resolve the URI using the requests api

        Args:
            url:
            id:
            context:

        Returns:

        """
        try:
            response = send_get_request(url)
            return self.resolve_string(response.content, context)
        except Exception as e:
            # if an error occurs return None to use the next registered resolver (or lxml default resolver)
            logger.error("An error occurred with the RequestsResolver. Please make sure HTTPS is properly configured.")
            return None



from core_main_app.utils.databases.mongoengine_database import Database

SECRET_KEY = 'fake-key'

INSTALLED_APPS = [
    # Django apps
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.sites',

    # Extra apps
    "defender",

    # Local apps
    "core_main_app",
    "tests",
]

# SERVER URI
SERVER_URI = "http://my.curator.com"

# IN-MEMORY TEST DATABASE
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    },
}

MIDDLEWARE = (
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            'templates'
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                "core_main_app.utils.custom_context_processors.domain_context_processor",  # Needed by any curator app
                "django.template.context_processors.i18n",
            ],
        },
    },
]

PASSWORD_HASHERS = (
    'django.contrib.auth.hashers.UnsaltedMD5PasswordHasher',
)

MOCK_DATABASE_NAME = 'db_mock'
MOCK_DATABASE_HOST = 'mongomock://localhost'

DATA_SORTING_FIELDS = ['+title']

# Password settings for django.contrib.auth validators
# Specifies the minimum length for passwords.
PASSWORD_MIN_LENGTH = 12
# Specifies the minimum amount of required letters in a password.
PASSWORD_MIN_LETTERS = 1
# Specifies the minimum amount of required uppercase letters in a password.
PASSWORD_MIN_UPPERCASE_LETTERS = 1
# Specifies the minimum amount of required lowercase letters in a password.
PASSWORD_MIN_LOWERCASE_LETTERS = 1
# Specifies the minimum amount of required numbers in a password.
PASSWORD_MIN_NUMBERS = 1
# Specifies the minimum amount of required symbols in a password.
PASSWORD_MIN_SYMBOLS = 1
# Specifies the maximum amount of consecutive characters allowed in passwords.
PASSWORD_MAX_OCCURRENCE = 5

database = Database(MOCK_DATABASE_HOST, MOCK_DATABASE_NAME)
database.connect()
